r"""
The :py:mod:`vc2_conformance.test_cases` module contains routines for
generating test cases for VC-2 encoder and decoder implementations.

A description of each test case generator can be found in
:ref:`encoder-test-cases` and :ref:`decoder-test-cases` in the user guide.

Test case generators
--------------------

Test case generators are generator functions which take a
:py:class:`~vc2_conformance.codec_features.CodecFeatures` dictionary as their
only argument and produce picture sequences (for encoders) or bitstreams (for
decoders). Test case generators may take one of the following forms:

* Function which returns a single test case
* Function which returns a single test case or None (indicating no test case
  was produced)
* Generator function which yields multiple test cases.

A test case can be one of:

* For encoder test cases, a picture sequence in the form of an
  :py:class:`EncoderTestSequence` object.
* For decoder test cases, a bitstream in the form of a
  :py:class:`~vc2_conformance.bitstream.Stream` object which can be serialised
  using
  :py:func:`~vc2_conformance.bitstream.vc2_autofill.autofill_and_serialise_stream`.
* A :py:class:`TestCase` object containing one of the above as its
  :py:attr:`~TestCase.value`.

Test case generators may prefer to output :py:class:`TestCase` objects when
multiple test cases are produced so that each testcase can be given its own
'subcase' name. In addition, test cases may be accompanied by a freeform JSON
serialisable metadata object when :py:class:`TestCase` objects are produced.

.. autoclass:: TestCase
    :members:
    :undoc-members:

All test case generator functions are run via
:py:func:`normalise_test_case_generator` which normalises the function into the
form of a generator which yields :py:class:`TestCase` objects. This also
populates the :py:attr:`~TestCase.case_name` field of the generated
:py:class:`TestCase` automatically with the test case generator's function
name.

.. autofunction:: normalise_test_case_generator


Encoder test case generators
----------------------------

Encoder test case generators are located in
:py:mod:`vc2_conformance.test_cases.encoder` must be decorated with the
:py:data:`vc2_conformance.test_cases.encoder_test_case_generator` decorator,
take a :py:class:`~vc2_conformance.codec_features.CodecFeatures` and produce
:py:class:`EncoderTestSequence` objects.

.. autodata:: encoder_test_case_generator
    :annotation:

.. autoclass:: EncoderTestSequence

Decoder test case generators
----------------------------

Decoder test case generators are located in
:py:mod:`vc2_conformance.test_cases.decoder` must be decorated with the
:py:data:`vc2_conformance.test_cases.decoder_test_case_generator` decorator,
take a :py:class:`~vc2_conformance.codec_features.CodecFeatures` and produce
:py:class:`vc2_conformance.bitstream.Stream` dictionaries which can be
serialised using
:py:func:`~vc2_conformance.bitstream.vc2_autofill.autofill_and_serialise_stream`.

.. autodata:: decoder_test_case_generator
    :annotation:


Test case generator registries
------------------------------

All test case generators are registered (by the
:py:data:`encoder_test_case_generator` and
:py:data:`decoder_test_case_generator` decorators) with one of two
:py:class:`Registry` singletons:

.. autodata:: ENCODER_TEST_CASE_GENERATOR_REGISTRY
    :annotation:

.. autodata:: DECODER_TEST_CASE_GENERATOR_REGISTRY
    :annotation:

The :py:class:`Registry` class implements a registry of test case generators
which the :ref:`vc2-test-case-generator` script uses to generate a complete set
of test cases.

.. autoclass:: Registry
    :members:

"""

from types import GeneratorType

from functools import partial

from collections import namedtuple


__all__ = [
    "EncoderTestSequence",
    "TestCase",
    "ENCODER_TEST_CASE_GENERATOR_REGISTRY",
    "DECODER_TEST_CASE_GENERATOR_REGISTRY",
]


EncoderTestSequence = namedtuple(
    "EncoderTestSequence", "pictures,video_parameters,picture_coding_mode"
)
"""
A sequence of pictures to be encoded by a VC-2 encoder under test.

Parameters
==========
pictures : [{"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}, ...]
    A :py:class:`list` of dictionaries containing raw picture data in 2D arrays
    for each picture component.
video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    The video parameters associated with the test sequence.
picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    The picture coding mode associated with the test sequence.
"""


class TestCase(object):
    """
    A test case, produced by a test case generator function.

    Parameters
    ==========
    value : :py:class:`EncoderTestSequence` or :py:class:`~vc2_conformance.bitstream.Stream`
        A value containing the test case itself. An
        :py:class:`EncoderTestSequence` for encoder test cases or
        :py:class:`~vc2_conformance.bitstream.Stream` for decoder test cases.
    subcase_name : str or None
        An identifier for the sub-case if this test case has several
        sub-cases.
    case_name : str or None
        The name of the test case. If None, the ``case_name`` will be
        populated automatically with the name of the test case generator
        function which returned it.
    metadata : object or None
        Optional JSON-serialisable metadata associated with this test case.
        The meaning and formatting of of this metadata is left to the
        individual test case to define.
    """

    def __init__(self, value, subcase_name=None, case_name=None, metadata=None):
        self._value = value
        self._subcase_name = subcase_name
        self._case_name = case_name
        self._metadata = metadata

    @property
    def name(self):
        """
        The complete name of this test case. Constructed from the
        :py:attr:`case_name` and  :py:attr:`subcase_name` attributes.

        A string of the form ``"case_name"`` or ``"case_name[subcase_name]"``.
        """
        if self.subcase_name is None:
            return self.case_name
        else:
            return "{}[{}]".format(self.case_name, self.subcase_name)

    @property
    def case_name(self):
        return self._case_name

    @case_name.setter
    def case_name(self, value):
        self._case_name = value

    @property
    def subcase_name(self):
        return self._subcase_name

    @subcase_name.setter
    def subcase_name(self, value):
        self._subcase_name = value

    @property
    def value(self):
        return self._value

    @property
    def metadata(self):
        return self._metadata

    def __repr__(self):
        return "<{} {}>".format(
            type(self).__name__,
            self.name,
        )

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.value == other.value
            and self.case_name == other.case_name
            and self.subcase_name == other.subcase_name
            and self.metadata == other.metadata
        )

    def __ne__(self, other):
        """Required under Python 2.x"""
        return not self.__eq__(other)


def normalise_test_case_generator(f, *args, **kwargs):
    """
    Call a test case generator, f, and, regardless of its native output,
    produces a generator of :py:class:`TestCase` objects.

    If the function returns or yields :py:class:`TestCase` objects, their
    :py:attr:`TestCase.case_name` attributes will be populated with the
    function name, if not already defined. If the function returns or generates
    other values, these will be wrapped in :py:class:`TestCase` objects
    automatically. If the function returns or generates None, no test case will
    be emitted.
    """
    generator = f(*args, **kwargs)

    is_generator = isinstance(generator, GeneratorType)
    if not is_generator:
        generator = [generator]

    for i, value in enumerate(generator):
        if value is None:
            continue

        if not isinstance(value, TestCase):
            value = TestCase(value)

        if value.case_name is None:
            value.case_name = f.__name__

        if is_generator and value.subcase_name is None:
            value.subcase_name = str(i)

        yield value


class Registry(object):
    """
    A registry of test case generating functions.
    """

    def __init__(self):
        self._test_case_generators = []

    def register_test_case_generator(self, f):
        """
        Register a test case generator function with this registry.

        Returns the (unmodified) function allowing this method to be used as a
        decorator.
        """
        self._test_case_generators.append(f)
        return f

    def generate_test_cases(self, *args, **kwargs):
        """
        Run every test case generator registered with this registry, passing
        each generator the supplied arguments. Generates all
        :py:class:`TestCase` objects.
        """
        for test_case_generator in self._test_case_generators:
            for test_case in normalise_test_case_generator(
                test_case_generator, *args, **kwargs
            ):
                yield test_case

    def iter_independent_generators(self, *args, **kwargs):
        """
        Produce a series of generator functions which may be called in
        parallel. Each returned zero-argument function should be called and
        will generate a series of :py:class:`TestCase` objects.
        """
        for test_case_generator in self._test_case_generators:
            yield partial(
                normalise_test_case_generator, test_case_generator, *args, **kwargs
            )

    def iter_registered_functions(self):
        """
        Iterates over the raw functions registered with this registry.

        Only intended for use during documentation generation.
        """
        for test_case_generator in self._test_case_generators:
            yield test_case_generator


ENCODER_TEST_CASE_GENERATOR_REGISTRY = Registry()
"""
:py:class:`Registry` singleton with which all VC-2 encoder test cases are
registered.
"""

encoder_test_case_generator = (
    ENCODER_TEST_CASE_GENERATOR_REGISTRY.register_test_case_generator
)
"""Decorator to use to register all encoder test case generators."""


DECODER_TEST_CASE_GENERATOR_REGISTRY = Registry()
"""
:py:class:`Registry` singleton with which all VC-2 decoder test cases are
registered.
"""

decoder_test_case_generator = (
    DECODER_TEST_CASE_GENERATOR_REGISTRY.register_test_case_generator
)
"""Decorator to use to register all decoder test case generators."""


# Import test case containing submodules to ensure they become registered with
# the above registries
from vc2_conformance.test_cases.encoder import *  # noqa: E402
from vc2_conformance.test_cases.decoder import *  # noqa: E402



r"""
The :py:mod:`vc2_conformance.encoder.pictures` module contains simple routines
for compressing pictures in a VC-2 bitstream.

The picture encoding behaviour used by the encoder is encapsulated by the
:py:func:`make_picture_data_units` function which turns a series of pictures
(given as raw pixel values) into a series of
:py:class:`DataUnits <vc2_conformance.bitstream.DataUnit>`:

.. autofunction:: make_picture_data_units


Encoding algorithm
------------------

Depending on the lossy/lossless coding mode chosen, one of two simple
algorithms is used.

Lossless mode
`````````````

In lossless mode, every slice's ``qindex`` will be set to 0 (no quantization)
and all transform coefficients will be coded verbatim (though trailing zeros
will be coded implicitly).

Slices will be sized as large as necessary, though as small as possible.

The smallest ``slice_size_scaler`` possible will be used for each coded picture
independently.

.. note::

    In principle, lossless modes may occasionally make use of quantization to
    achieve better compression. For example where all transform coefficients
    are a multiple of the same quantisation factor. This encoder, however, does
    not do this.

Lossy mode
``````````

In lossy mode the ``qindex`` for each slice is chosen on a slice-by-slice
basis. The encoder tests quantization indices starting at zero and stopping
when the transform coefficients fit into the slice.

Slices are sized such that the picture slice data in the bitstream totals
:py:class:`~vc2_conformance.codec_features.CodecFeatures`\
``["picture_bytes"]``.

For the high quality profile, the smallest ``slice_size_scaler`` which can
encode a slice where a single component consumes a whole slice is used for
every picture.

.. warning::

    The total size of picture slice data may differ from
    :py:class:`~vc2_conformance.codec_features.CodecFeatures`\
    ``["picture_bytes"]`` by up to ``slice_size_scaler`` bytes (for high
    quality profile formats) or one byte (for low delay profile formats). This
    will occur when the number of bytes (or multiple of ``slice_size_scaler``
    bytes) is not exactly divisible by the required number of picture bytes.

.. warning::

    The total number of bytes used to encode each picture, once other coding
    overheads (such as headers) will be higher than
    :py:class:`~vc2_conformance.codec_features.CodecFeatures`\
    ``["picture_bytes"]``.

.. note::

    This codec may not always produce highest quality pictures possible in
    lossy modes. For example, sometimes chosing higher quantisation indices can
    produce fewer coding artefacts, particularly in concatenated coding
    applications. Similarly, higher picture quality may sometimes be obtained
    by setting later transform coefficients to zero enabling a lower
    quantization index to be used. Other more sophisticated schemes may also
    directly tweak transform coefficients.


Use of pseudocode
-----------------

This module uses the pseudocode-derived
:py:mod:`vc2_conformance.pseudocode.picture_encoding` module for its
forward-DWT and :py:mod:`vc2_conformance.pseudocode.quantization` for
quantization. Other pseudocode routines are also used where possible, for
example for computing slice dimensions.

"""

from itertools import count

from collections import namedtuple

from fractions import Fraction

from copy import deepcopy

from vc2_data_tables import (
    QUANTISATION_MATRICES,
    Profiles,
    ParseCodes,
)

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.pseudocode.vc2_math import mean, intlog2

from vc2_conformance.pseudocode.arrays import width, height

from vc2_conformance.pseudocode.state import State

from vc2_conformance.pseudocode.video_parameters import set_coding_parameters

from vc2_conformance.pseudocode.picture_encoding import picture_encode

from vc2_conformance.codec_features import codec_features_to_trivial_level_constraints

from vc2_conformance.constraint_table import allowed_values_for, ValueSet

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.pseudocode.slice_sizes import (
    slice_bytes,
    slice_left,
    slice_right,
    slice_top,
    slice_bottom,
)

from vc2_conformance.pseudocode.quantization import forward_quant

from vc2_conformance.bitstream import (
    HQSlice,
    LDSlice,
    TransformData,
    QuantMatrix,
    PictureHeader,
    SliceParameters,
    TransformParameters,
    ExtendedTransformParameters,
    WaveletTransform,
    PictureParse,
    DataUnit,
    ParseInfo,
    FragmentParse,
    FragmentHeader,
    FragmentData,
)

from vc2_conformance.bitstream.exp_golomb import signed_exp_golomb_length

from vc2_conformance.encoder.exceptions import (
    MissingQuantizationMatrixError,
    IncompatibleLevelAndExtendedTransformParametersError,
    PictureBytesSpecifiedForLosslessModeError,
    InsufficientHQPictureBytesError,
    InsufficientLDPictureBytesError,
    LosslessUnsupportedByLowDelayError,
)


def get_quantization_marix(codec_features):
    """
    Get the quantization matrix (in the form of the ``quant_matrix`` entry in
    ``state``) based on the provided ``codec_features``.

    Throws
    :py:exc:`vc2_conformance.encoder.exceptions.MissingQuantizationMatrixError`
    if a matrix is not provided when no default is available.
    """
    if codec_features["quantization_matrix"] is None:
        transform_details = (
            codec_features["wavelet_index"],
            codec_features["wavelet_index_ho"],
            codec_features["dwt_depth"],
            codec_features["dwt_depth_ho"],
        )
        if transform_details not in QUANTISATION_MATRICES:
            raise MissingQuantizationMatrixError(codec_features)
        return QUANTISATION_MATRICES[transform_details]
    else:
        return codec_features["quantization_matrix"]


def serialize_quantization_matrix(matrix):
    """
    Given a quantization matrix as a hierarchy of dictionaries, returns the
    serial list of matrix values for use in a VC-2 bitstream.
    """
    out = []
    for level, orients in sorted(matrix.items()):
        for orient, value in sorted(
            orients.items(),
            key=lambda ov: ["L", "H", "LL", "HL", "LH", "HH"].index(ov[0]),
        ):
            out.append(value)
    return out


@ref_pseudocode(deviation="inferred_implementation")
def apply_dc_prediction(band):
    """(13.4) Inverse of dc_prediction."""
    for y in reversed(range(0, height(band))):
        for x in reversed(range(0, width(band))):
            if x > 0 and y > 0:
                prediction = mean(band[y][x - 1], band[y - 1][x - 1], band[y - 1][x])
            elif x > 0 and y == 0:
                prediction = band[0][x - 1]
            elif x == 0 and y > 0:
                prediction = band[y - 1][0]
            else:
                prediction = 0
            band[y][x] -= prediction


def calculate_coeffs_bits(coeffs):
    """
    Calculate the number of bits required to represent the supplied sequence of
    coefficients using signed exp-golomb coding within a bounded block.

    Parameters
    ==========
    coeffs : [int, ...]

    Returns
    =======
    num_bits : int
    """
    num_bits = 0

    skip_zeros = True
    for coeff in reversed(coeffs):
        if skip_zeros and coeff == 0:
            continue
        else:
            skip_zeros = False
            num_bits += signed_exp_golomb_length(coeff)

    return num_bits


def calculate_hq_length_field(coeffs, slice_size_scaler):
    """
    Compute a HQ picture slice length field for a set of coefficients.
    """
    multiple = 8 * slice_size_scaler
    return (calculate_coeffs_bits(coeffs) + multiple - 1) // multiple


def quantize_coeffs(qindex, coeff_values, quant_matrix_values):
    """
    Quantize a set of coefficient values.

    Parameters
    ==========
    qindex : int
        The base quantization index.
    coeff_values : [int, ...]
        The coefficients to be quantized.
    quant_matrix_values : [int, ...]
        For each entry in ``coeff_value``, the corresponding quantization
        matrix value.

    Returns
    =======
    quantized_coeff_values : [int, ...]
    """
    return [
        forward_quant(
            coeff_value,
            max(0, qindex - quant_matrix_value),
        )
        for coeff_value, quant_matrix_value in zip(
            coeff_values,
            quant_matrix_values,
        )
    ]


ComponentCoeffs = namedtuple("ComponentCoeffs", "coeff_values,quant_matrix_values")
"""
A tuple containing (in bitstream order) the transform coefficients and
corresponding quantisation matrix values for a particular picture component
within a picture slice.
"""

SliceCoeffs = namedtuple("SliceCoeffs", "Y,C1,C2")
"""
A tuple of :py:class:`ComponentCoeffs` tuples giving the transform coefficients
described by a picture slice.
"""


def quantize_to_fit(target_size, coeff_sets, align_bits=1, minimum_qindex=0):
    """
    Find the quantisation index necessary to reduce a several sets of transform
    coefficients to total a target length in bits.

    Each block of quantized transform coefficients is assumed to be padded to a
    whole multiple of align_bits bits.

    Parameters
    ==========
    target_size : int
        Target size, in bytes.
    coeff_sets : [:py:class:`ComponentCoeffs`, ...]
        A series of sets of coeffs to be quantised and concatenated, along with
        their corresponding quantisation matrix values.
    align_bits : int
        The size of each block will be rounded up to a whole multiple of this
        value. Use this to, for example, force lengths to be a whole number of
        bytes (set to 8) or some slice_size_scaler (set to
        8*slice_size_scaler).
    minimum_qindex : int
        If provided, gives the quantization index to start with when trying to
        find a suitable quantization index.

    Returns
    =======
    qindex : int
        The quantisation index chosen.
    [[int, ...], ...]
        For each set of coefficients provided, the quantised coefficient
        values.
    """
    assert target_size >= 0

    for qindex in count(minimum_qindex):
        quantized_coeff_sets = [
            quantize_coeffs(
                qindex,
                component_coeffs.coeff_values,
                component_coeffs.quant_matrix_values,
            )
            for component_coeffs in coeff_sets
        ]

        total_length = sum(
            # Round each block's length to whole multiple of align_bits
            ((calculate_coeffs_bits(quantized_coeffs) + align_bits - 1) // align_bits)
            * align_bits
            for quantized_coeffs in quantized_coeff_sets
        )

        if total_length <= target_size:
            return (qindex, quantized_coeff_sets)


def transform_and_slice_picture(codec_features, picture):
    """
    Transform a picture provided using a forward DWT and DC prediction
    according to the :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    dictionary provided. Returns transform coefficients grouped by picture
    slice.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
        The picture to be encoded. This picture will be compressed using a
        simple VC-2 encoder implementation. It does not necessarily produce the
        most high-quality encodings.

    Returns
    =======
    slice_coeffs : [[:py:class:`SliceCoeffs`, ...], ...]
        A 2D array containing, for each picture slice, the transform
        coefficients and corresponding quantisation matrix values in bitstream
        order.
    """
    # Perform a forward DWT
    state = State(
        wavelet_index=codec_features["wavelet_index"],
        wavelet_index_ho=codec_features["wavelet_index_ho"],
        dwt_depth=codec_features["dwt_depth"],
        dwt_depth_ho=codec_features["dwt_depth_ho"],
        slices_x=codec_features["slices_x"],
        slices_y=codec_features["slices_y"],
        picture_coding_mode=codec_features["picture_coding_mode"],
    )
    set_coding_parameters(state, codec_features["video_parameters"])

    # NB: picture_encode corrupts the supplied picture arrays so a copy is
    # provided here
    picture_encode(state, deepcopy(picture))

    # Perform DC prediction
    if codec_features["profile"] == Profiles.low_delay:
        if state["dwt_depth_ho"] == 0:
            apply_dc_prediction(state["y_transform"][0]["LL"])
            apply_dc_prediction(state["c1_transform"][0]["LL"])
            apply_dc_prediction(state["c2_transform"][0]["LL"])
        else:
            apply_dc_prediction(state["y_transform"][0]["L"])
            apply_dc_prediction(state["c1_transform"][0]["L"])
            apply_dc_prediction(state["c2_transform"][0]["L"])

    # Load quantisation matrix
    state["quant_matrix"] = get_quantization_marix(codec_features)

    # Divide the picture into slices and collect together transform
    # coefficients in bitstream order (along with associated quantisation
    # matrix values)
    slice_coeffs = [
        [
            SliceCoeffs(
                ComponentCoeffs([], []),
                ComponentCoeffs([], []),
                ComponentCoeffs([], []),
            )
            for _ in range(state["slices_x"])
        ]
        for _ in range(state["slices_y"])
    ]

    # NB: Iteration order for level and orient are critical here
    for transform in ["y_transform", "c1_transform", "c2_transform"]:
        comp = transform.split("_")[0].upper()
        for level, orients in sorted(state[transform].items()):
            for orient, coeffs in sorted(
                orients.items(),
                key=lambda orient_coeffs: ["L", "LL", "H", "HL", "LH", "HH"].index(
                    orient_coeffs[0]
                ),
            ):

                sxs = [
                    (
                        slice_left(state, sx, comp, level),
                        slice_right(state, sx, comp, level),
                    )
                    for sx in range(state["slices_x"])
                ]
                sys = [
                    (
                        slice_top(state, sy, comp, level),
                        slice_bottom(state, sy, comp, level),
                    )
                    for sy in range(state["slices_y"])
                ]

                for sy, (y1, y2) in enumerate(sys):
                    for sx, (x1, x2) in enumerate(sxs):
                        for y in range(y1, y2):
                            for x in range(x1, x2):
                                sc = getattr(slice_coeffs[sy][sx], comp)
                                sc.coeff_values.append(coeffs[y][x])
                                sc.quant_matrix_values.append(
                                    state["quant_matrix"][level][orient]
                                )

    return slice_coeffs


def make_hq_slice(
    y_transform,
    c1_transform,
    c2_transform,
    total_length,
    qindex,
    slice_size_scaler=1,
):
    """
    Create a :py:class:`vc2_conformance.bitstream.HQSlice` containing the
    provided already-quantized coefficients.

    Parameters
    ==========
    y_transform : [int, ...]
    c1_transform : [int, ...]
    c2_transform : [int, ...]
        Quantized transform coefficients for this slice, in bitstream order.
    total_length : int or None
        If an integer, the ``slice_c2_length`` field will be set such that all
        three length fields sum to ``total_length``. Note that this length
        value is given in multiples of slice_size_scaler bytes and that this
        total length does not account for the size of the length fields and
        qindex fields also present in the slice.

        If None, the ``slice_c2_length`` field will be set to the minimum
        multiple of slice_size_scaler bytes which the ``c2_transform``
        coefficients can fit into.
    qindex : int
        The quantization index used to quantise the provided coefficient.s
    slice_size_scaler : int
    """
    # Compute component lengths
    y_length = calculate_hq_length_field(y_transform, slice_size_scaler)
    c1_length = calculate_hq_length_field(c1_transform, slice_size_scaler)
    if total_length is None:
        c2_length = calculate_hq_length_field(c2_transform, slice_size_scaler)
    else:
        c2_length = total_length - y_length - c1_length

    return HQSlice(
        qindex=qindex,
        slice_y_length=y_length,
        slice_c1_length=c1_length,
        slice_c2_length=c2_length,
        y_transform=y_transform,
        c1_transform=c1_transform,
        c2_transform=c2_transform,
    )


def make_ld_slice(y_transform, c_transform, qindex):
    """
    Create a :py:class:`vc2_conformance.bitstream.LDSlice` containing the
    provided already-quantized coefficients.

    Parameters
    ==========
    y_transform : [int, ...]
    c_transform : [int, ...]
        Quantized transform coefficients for this slice, in bitstream order.
    qindex : int
        The quantization index used to quantise the provided coefficient.s
    """
    return LDSlice(
        qindex=qindex,
        slice_y_length=calculate_coeffs_bits(y_transform),
        y_transform=y_transform,
        c_transform=c_transform,
    )


def make_transform_data_hq_lossless(transform_coeffs, minimum_slice_size_scaler=1):
    """
    Pack transform coefficients into HQ picture slices in a
    :py:class:`TransformData`, computing the required slice_size_scaler.

    Parameters
    ==========
    transform_coeffs : [[:py:class:`SliceCoeffs`, ...], ...]
    minimum_slice_size_scaler : int
        Specifies the minimum slice_size_scaler to be used for high quality
        pictures. Ignored in low delay mode.

    Returns
    =======
    slice_size_scaler : int
    transform_data : :py:class:`vc2_conformance.bitstream.TransformData`
    """
    # Initially, the slice_size_scaler is assumed to be 1. If any of the length
    # fields which result are > 255 we will need to revise this, along with all
    # of the length values.
    transform_data = TransformData(
        hq_slices=[
            make_hq_slice(
                transform_coeffs_slice.Y.coeff_values,
                transform_coeffs_slice.C1.coeff_values,
                transform_coeffs_slice.C2.coeff_values,
                total_length=None,
                qindex=0,
                slice_size_scaler=1,
            )
            for transform_coeffs_row in transform_coeffs
            for transform_coeffs_slice in transform_coeffs_row
        ]
    )

    # Find the minimum slice size scaler possible
    max_length = max(
        max(
            hq_slice["slice_y_length"],
            hq_slice["slice_c1_length"],
            hq_slice["slice_c2_length"],
        )
        for hq_slice in transform_data["hq_slices"]
    )
    slice_size_scaler = max(1, minimum_slice_size_scaler, (max_length + 254) // 255)

    # Adjust slice lengths accordingly
    for hq_slice in transform_data["hq_slices"]:
        hq_slice["slice_y_length"] += slice_size_scaler - 1
        hq_slice["slice_y_length"] //= slice_size_scaler

        hq_slice["slice_c1_length"] += slice_size_scaler - 1
        hq_slice["slice_c1_length"] //= slice_size_scaler

        hq_slice["slice_c2_length"] += slice_size_scaler - 1
        hq_slice["slice_c2_length"] //= slice_size_scaler

    return slice_size_scaler, transform_data


def get_safe_lossy_hq_slice_size_scaler(picture_bytes, num_slices):
    """
    Return a slice_size_scaler which is guaranteed to be large enough for any
    slice in a lossy HQ picture.

    Parameters
    ==========
    picture_bytes : int
        The total number of bytes the picture slices should take up (i.e.
        including ``qindex`` and ``slice_{y,c1,c2}_length`` fields). Must allow
        at least 4 bytes per slice.
    num_slices : int
        Number of slices per picture.

    Returns
    =======
    slice_size_scaler : int
        A safe slice size scaler to use.
    """
    max_slice_bytes = (picture_bytes + num_slices - 1) // num_slices
    max_length_field_value = max_slice_bytes - 4

    # Find the smallest slice size scaler which lets this size fit into an 8
    # bit length field.
    slice_size_scaler = (max_length_field_value + 254) // 255

    return max(1, slice_size_scaler)


def make_transform_data_hq_lossy(
    picture_bytes, transform_coeffs, minimum_qindex=0, minimum_slice_size_scaler=1
):
    """
    Quantize and pack transform coefficients into HQ picture slices in a
    :py:class:`TransformData`.

    Raises :py:exc:`InsufficientHQPictureBytesError` if ``picture_bytes`` is too
    small.

    Parameters
    ==========
    picture_bytes : int
        The total number of bytes the picture slices should take up (i.e.
        including ``qindex`` and ``slice_{y,c1,c2}_length`` fields). Must allow
        at least 4 bytes per slice. When slice sizes are large enough to
        require a slice_size_scaler larger than 1, ``picture_bytes -
        4*num_slices`` must be a multiple of ``slice_size_scaler``, otherwise
        the total picture size will deviate by up to ``slice_size_scaler``
        bytes from ``picture_bytes``.
    transform_coeffs : [[:py:class:`SliceCoeffs`, ...], ...]
    minimum_qindex : int
        If provided, gives the quantization index to start with when trying to
        find a suitable quantization index.
    minimum_slice_size_scaler : int
        Specifies the minimum slice_size_scaler to be used for high quality
        pictures. Ignored in low delay mode.

    Returns
    =======
    slice_size_scaler : int
    transform_data : :py:class:`vc2_conformance.bitstream.TransformData`
    """
    slices_x = width(transform_coeffs)
    slices_y = height(transform_coeffs)

    # Determine the largest size a slice could be
    num_slices = slices_x * slices_y

    # Find the smallest slice size scaler which lets this size fit into 8 bits
    slice_size_scaler = max(
        get_safe_lossy_hq_slice_size_scaler(picture_bytes, num_slices),
        minimum_slice_size_scaler,
    )

    # Work out the total bitstream space available after slice overheads are
    # accounted for (NB: 4 bytes overhead per slice due to qindex and
    # slice_{y,c1,c2}_length fields).
    total_coeff_bytes = picture_bytes - (num_slices * 4)

    if total_coeff_bytes < 0:
        raise InsufficientHQPictureBytesError()

    # We'll repurpose slice_bytes (13.5.3.2) to compute the number of
    # slice_size_scaler bytes available for transform coefficients in each
    # picture slice. (i.e. it won't compute the size of the slice in bytes but
    # the number of slice_size_scaler bytes available for transform
    # coefficients in each slice).
    state = State(
        slices_x=slices_x,
        slices_y=slices_y,
        slice_bytes_numerator=total_coeff_bytes,
        slice_bytes_denominator=num_slices * slice_size_scaler,
    )

    transform_data = TransformData(hq_slices=[])
    for sy, transform_coeffs_row in enumerate(transform_coeffs):
        for sx, transform_coeffs_slice in enumerate(transform_coeffs_row):
            # NB: Actually calculates multiples of slice_size_scaler bytes
            # after all length/qindex fields accounted for. See comment above
            # "state = State(".
            total_length = slice_bytes(state, sx, sy)
            target_size = 8 * slice_size_scaler * total_length

            # Quantize each slice to fit
            qindex, (y_transform, c1_transform, c2_transform) = quantize_to_fit(
                target_size,
                transform_coeffs_slice,
                8 * slice_size_scaler,
                minimum_qindex,
            )
            transform_data["hq_slices"].append(
                make_hq_slice(
                    y_transform,
                    c1_transform,
                    c2_transform,
                    total_length,
                    qindex,
                    slice_size_scaler,
                )
            )

    return slice_size_scaler, transform_data


def interleave(a, b):
    """
    Return a list containing the interleaving of elements from the lists a and
    b, first element from 'a' first.
    """
    out = []
    for va, vb in zip(a, b):
        out.append(va)
        out.append(vb)
    return out


def make_transform_data_ld_lossy(picture_bytes, transform_coeffs, minimum_qindex=0):
    """
    Quantize and pack transform coefficients into LD picture slices in a
    :py:class:`TransformData`.

    Raises :py:exc:`InsufficientLDPictureBytesError` if ``picture_bytes`` is too
    small.

    Parameters
    ==========
    picture_bytes : int
        The total number of bytes the picture slices should take up (i.e.
        including qindex and slice_y_length fields). Must be at least 1 byte
        per slice.
    transform_coeffs : [[:py:class:`SliceCoeffs`, ...], ...]
    minimum_qindex : int
        If provided, gives the quantization index to start with when trying to
        find a suitable quantization index.

    Returns
    =======
    transform_data : :py:class:`vc2_conformance.bitstream.TransformData`
    """
    # Used by slice_bytes
    state = State(
        slices_x=width(transform_coeffs),
        slices_y=height(transform_coeffs),
        slice_bytes_numerator=picture_bytes,
        slice_bytes_denominator=width(transform_coeffs) * height(transform_coeffs),
    )

    transform_data = TransformData(ld_slices=[])
    for sy, transform_coeffs_row in enumerate(transform_coeffs):
        for sx, transform_coeffs_slice in enumerate(transform_coeffs_row):
            target_size = 8 * slice_bytes(state, sx, sy)
            target_size -= 7  # qindex field
            target_size -= intlog2(target_size)  # slice_y_length field

            if target_size < 0:
                raise InsufficientLDPictureBytesError()

            # Interleave color components
            y_coeffs = transform_coeffs_slice.Y
            c_coeffs = ComponentCoeffs(
                coeff_values=interleave(
                    transform_coeffs_slice.C1.coeff_values,
                    transform_coeffs_slice.C2.coeff_values,
                ),
                quant_matrix_values=interleave(
                    transform_coeffs_slice.C1.quant_matrix_values,
                    transform_coeffs_slice.C2.quant_matrix_values,
                ),
            )

            # Quantize each slice to fit
            qindex, (y_transform, c_transform) = quantize_to_fit(
                target_size,
                [y_coeffs, c_coeffs],
                minimum_qindex=minimum_qindex,
            )
            transform_data["ld_slices"].append(
                make_ld_slice(
                    y_transform,
                    c_transform,
                    qindex,
                )
            )

    return transform_data


def make_quant_matrix(codec_features):
    """
    Create a :py:class:`vc2_conformance.bitstream.QuantMatrix` given a set of
    codec features.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`

    Returns
    =======
    quant_matrix : :py:class:`vc2_conformance.bitstream.QuantMatrix`
    """
    if codec_features["quantization_matrix"] is None:
        return QuantMatrix(custom_quant_matrix=False)
    else:
        return QuantMatrix(
            custom_quant_matrix=True,
            quant_matrix=serialize_quantization_matrix(
                codec_features["quantization_matrix"]
            ),
        )


def decide_extended_transform_flag(codec_features, flag_name, required):
    """
    Decide what asym_transform*_flag setting to use, accounting for level
    restrictions.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    flag_name : str
        The name of the flag to be decided (e.g. "asym_transform_index_flag").
    required : str
        If True, require this flag to be True, if False, the flag may be set to
        True or False, as allowed by the level, preferring False.

    Returns
    =======
    flag : bool

    Raises
    ======
    vc2_conformance.encoder.exceptions.IncompatibleLevelAndExtendedTransformParametersError
        If ``required`` is True but the level prohibits the flag.
    """
    # The flag states which we could use to encode the required value
    usable_flags = [True] if required else [False, True]

    # The allowed flag values according to the current level
    constrained_values = codec_features_to_trivial_level_constraints(codec_features)
    permitted_flags = allowed_values_for(
        LEVEL_CONSTRAINTS, flag_name, constrained_values
    )

    # Warning: Slight bodge/special case handled here...
    #
    # The make_extended_transform_parameters function (and by extension, this
    # function) is required to return an ExtendedTransformParameters object
    # even for streams with major_version < 3 which don't contain that field.
    #
    # To handle this case, we treat an empty constraint table entry as allowing
    # 'False' as a valid flag option when in reality for < v3 streams no value
    # is permitted.
    #
    # Note that we will still raise
    # IncompatibleLevelAndExtendedTransformParametersError if required is True
    # in this case. We also don't allow False if the level specifically
    # requires the flag to be True (i.e. we only allow it when the level gives
    # no values).
    if permitted_flags == ValueSet():
        permitted_flags.add_value(False)

    try:
        return next(flag for flag in usable_flags if flag in permitted_flags)
    except StopIteration:
        raise IncompatibleLevelAndExtendedTransformParametersError(codec_features)


def make_extended_transform_parameters(codec_features):
    """
    Create a :py:class:`vc2_conformance.bitstream.ExtendedTransformParameters`
    given a set of codec features. The encoding used will be as short as
    possible for the specified codec configuration.
    """

    etp = ExtendedTransformParameters()

    etp["asym_transform_index_flag"] = decide_extended_transform_flag(
        codec_features,
        "asym_transform_index_flag",
        codec_features["wavelet_index"] != codec_features["wavelet_index_ho"],
    )
    if etp["asym_transform_index_flag"]:
        etp["wavelet_index_ho"] = codec_features["wavelet_index_ho"]

    etp["asym_transform_flag"] = decide_extended_transform_flag(
        codec_features,
        "asym_transform_flag",
        codec_features["dwt_depth_ho"] != 0,
    )
    if etp["asym_transform_flag"]:
        etp["dwt_depth_ho"] = codec_features["dwt_depth_ho"]

    return etp


def make_picture_parse(
    codec_features, picture, minimum_qindex=0, minimum_slice_size_scaler=1
):
    """
    Compress a picture.

    Raises :py:exc:`PictureBytesSpecifiedForLosslessModeError` if
    ``picture_bytes`` is specifiied for a lossless coding mode.

    Raises :py:exc:`InsufficientLDPictureBytesError`
    :py:exc:`InsufficientHQPictureBytesError` if ``picture_bytes`` is too small
    for the coding mode used.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
        The picture to be encoded. This picture will be compressed using a
        simple VC-2 encoder implementation. It does not necessarily produce the
        most high-quality encodings. If ``pic_num`` is omitted,
        ``picture_number`` fields will be omitted in the output.
    minimum_qindex : int
        Specifies the minimum quantization index to be used. Must be 0 for
        lossless codecs.
    minimum_slice_size_scaler : int
        Specifies the minimum slice_size_scaler to be used for high quality
        pictures. Ignored in low delay mode.

    Returns
    =======
    transform_data : :py:class:`vc2_conformance.bitstream.TransformData`
        The transform data, ready for serialization.
    """
    # Apply transform and split into slices
    transform_coeffs = transform_and_slice_picture(codec_features, picture)

    picture_header = PictureHeader()
    if "pic_num" in picture:
        picture_header["picture_number"] = picture["pic_num"]

    slice_parameters = SliceParameters(
        slices_x=codec_features["slices_x"],
        slices_y=codec_features["slices_y"],
    )

    # Quantize coefficients as required and set slice_size_scaler (HQ Only) and
    # slice_bytes (LD only)
    if codec_features["profile"] == Profiles.high_quality:
        if codec_features["lossless"]:
            assert minimum_qindex == 0
            if codec_features["picture_bytes"] is not None:
                raise PictureBytesSpecifiedForLosslessModeError(codec_features)
            slice_size_scaler, transform_data = make_transform_data_hq_lossless(
                transform_coeffs,
                minimum_slice_size_scaler,
            )
        else:
            try:
                slice_size_scaler, transform_data = make_transform_data_hq_lossy(
                    codec_features["picture_bytes"],
                    transform_coeffs,
                    minimum_qindex,
                    minimum_slice_size_scaler,
                )
            except InsufficientHQPictureBytesError:
                # Re-raise with codec features dict
                raise InsufficientHQPictureBytesError(codec_features)

        # NB: For simplicity, this implementation currently does not support
        # setting the slice prefix bytes to anything except zero since this is
        # not required by any existing VC-2 level. The assumption that this is
        # OK is verified in
        # ``tests/encoder/test_level_constraints_assumptions.py``.
        #
        # In addition, the `codec_features_to_trivial_level_constraints`
        # function assumes that this value is always 0 too.
        slice_parameters["slice_prefix_bytes"] = 0
        slice_parameters["slice_size_scaler"] = slice_size_scaler
    elif codec_features["profile"] == Profiles.low_delay:
        if codec_features["lossless"]:
            raise LosslessUnsupportedByLowDelayError(codec_features)
        try:
            transform_data = make_transform_data_ld_lossy(
                codec_features["picture_bytes"],
                transform_coeffs,
                minimum_qindex,
            )
        except InsufficientLDPictureBytesError:
            # Re-raise with codec features dict
            raise InsufficientLDPictureBytesError(codec_features)

        slice_bytes_fraction = Fraction(
            codec_features["picture_bytes"],
            codec_features["slices_x"] * codec_features["slices_y"],
        )
        slice_parameters["slice_bytes_numerator"] = slice_bytes_fraction.numerator
        slice_parameters["slice_bytes_denominator"] = slice_bytes_fraction.denominator

    transform_parameters = TransformParameters(
        wavelet_index=codec_features["wavelet_index"],
        dwt_depth=codec_features["dwt_depth"],
        # NB: We *always* include an ExtendedTransformParameters field. This
        # field will later be removed (if not supported by the major_version
        # chosen for the stream) using the
        # :py:func:`vc2_conformance.bitstream.vc2_autofill.autofill_major_version`
        # function.
        extended_transform_parameters=make_extended_transform_parameters(
            codec_features
        ),
        slice_parameters=slice_parameters,
        quant_matrix=make_quant_matrix(codec_features),
    )

    wavelet_transform = WaveletTransform(
        transform_parameters=transform_parameters,
        transform_data=transform_data,
    )

    return PictureParse(
        picture_header=picture_header,
        wavelet_transform=wavelet_transform,
    )


def make_picture_parse_data_unit(
    codec_features, picture, minimum_qindex=0, minimum_slice_size_scaler=1
):
    """
    Create a :py:class:`~vc2_conformance.bitstream.DataUnit` object containing
    a (possibly lossily compressed) picture.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
        The picture to be encoded. This picture will be compressed using a
        simple VC-2 encoder implementation. It does not necessarily produce the
        most high-quality encodings. If ``pic_num`` is omitted,
        ``picture_number`` fields will be omitted in the output.
    minimum_qindex : int
        Specifies the minimum quantization index to be used. Must be 0 for
        lossless codecs.
    minimum_slice_size_scaler : int
        Specifies the minimum slice_size_scaler to be used for high quality
        pictures. Ignored in low delay mode.

    Returns
    =======
    data_unit : :py:class:`vc2_conformance.bitstream.DataUnit`
    """
    assert codec_features["fragment_slice_count"] == 0

    return DataUnit(
        parse_info=ParseInfo(
            parse_code=(
                ParseCodes.high_quality_picture
                if codec_features["profile"] == Profiles.high_quality
                else ParseCodes.low_delay_picture
                if codec_features["profile"] == Profiles.low_delay
                else None  # Unreachable, unless a new profile is added
            )
        ),
        picture_parse=make_picture_parse(
            codec_features, picture, minimum_qindex, minimum_slice_size_scaler
        ),
    )


def make_fragment_parse_data_units(
    codec_features, picture, minimum_qindex=0, minimum_slice_size_scaler=1
):
    r"""
    Create a series of :py:class:`DataUnits
    <vc2_conformance.bitstream.DataUnit>` encoding a (possibly lossily
    compressed) picture.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
        The picture to be encoded. This picture will be compressed using a
        simple VC-2 encoder implementation. It does not necessarily produce the
        most high-quality encodings. If ``pic_num`` is omitted,
        ``picture_number`` fields will be omitted in the output.
    minimum_qindex : int
        Specifies the minimum quantization index to be used. Must be 0 for
        lossless codecs.
    minimum_slice_size_scaler : int
        Specifies the minimum slice_size_scaler to be used for high quality
        pictures. Ignored in low delay mode.

    Returns
    =======
    fragment_data_units : [:py:class:`vc2_conformance.bitstream.DataUnit`, ...]
    """
    assert codec_features["fragment_slice_count"] != 0

    # To avoid repeating ourselves, the fragmented picture is assembled from
    # the parts of a ready-made piture_parse.
    picture_parse = make_picture_parse(
        codec_features, picture, minimum_qindex, minimum_slice_size_scaler
    )

    wavelet_transform = picture_parse["wavelet_transform"]

    transform_parameters = wavelet_transform["transform_parameters"]
    transform_data = wavelet_transform["transform_data"]

    if codec_features["profile"] == Profiles.high_quality:
        parse_code = ParseCodes.high_quality_picture_fragment
        slices_name = "hq_slices"
    elif codec_features["profile"] == Profiles.low_delay:
        parse_code = ParseCodes.low_delay_picture_fragment
        slices_name = "ld_slices"

    fragment_data_units = []

    # Add the first fragment containing the transform parameters
    fragment_data_units.append(
        DataUnit(
            parse_info=ParseInfo(parse_code=parse_code),
            fragment_parse=FragmentParse(
                fragment_header=FragmentHeader(
                    fragment_data_length=0,
                    fragment_slice_count=0,
                ),
                transform_parameters=transform_parameters,
            ),
        )
    )

    # A count of how many slices worth of space remain in the current
    # (slice-containing) fragment. Initially set to zero as we don't have any
    # picture containing fragments.
    fragment_slices_remaining = 0

    # Add the remaining fragments containing the picture slices
    slice_iterator = iter(transform_data[slices_name])
    for sy in range(codec_features["slices_y"]):
        for sx in range(codec_features["slices_x"]):
            # If the current fragment is full, start a new one
            if fragment_slices_remaining == 0:
                fragment_slices_remaining = codec_features["fragment_slice_count"]
                fragment_data_units.append(
                    DataUnit(
                        parse_info=ParseInfo(parse_code=parse_code),
                        fragment_parse=FragmentParse(
                            fragment_header=FragmentHeader(
                                fragment_data_length=0,
                                # NB: Will be incremented in the next step(s)
                                fragment_slice_count=0,
                                fragment_x_offset=sx,
                                fragment_y_offset=sy,
                            ),
                            fragment_data=FragmentData(
                                {
                                    # NB: Will be populated in the next step(s)
                                    slices_name: [],
                                }
                            ),
                        ),
                    )
                )

            # Add the slice to the current fragment
            fragment_parse = fragment_data_units[-1]["fragment_parse"]
            fragment_parse["fragment_header"]["fragment_slice_count"] += 1
            fragment_parse["fragment_data"][slices_name].append(next(slice_iterator))
            fragment_slices_remaining -= 1

    # Populate picture_number field in fragment headers, if one is provided
    if "pic_num" in picture:
        for data_unit in fragment_data_units:
            fragment_header = data_unit["fragment_parse"]["fragment_header"]
            fragment_header["picture_number"] = picture["pic_num"]

    return fragment_data_units


def make_picture_data_units(
    codec_features,
    picture,
    minimum_qindex=0,
    minimum_slice_size_scaler=1,
):
    r"""
    Create a seires of one or more :py:class:`DataUnits
    <vc2_conformance.bitstream.DataUnit>` containing a compressed version of
    the supplied picture.

    When ``codec_features["fragment_slice_count"]`` is 0, a single picture
    parse data unit will be produced. otherwise a series of two or more
    fragment parse data units will be produced.

    A simple wrapper around :py:func:`make_picture_parse_data_unit` and
    :py:func:`make_fragment_parse_data_units`.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
        The picture to be encoded. This picture will be compressed using a
        simple VC-2 encoder implementation. It does not necessarily produce the
        most high-quality encodings. If ``pic_num`` is omitted,
        ``picture_number`` fields will be omitted in the output.
    minimum_qindex : int
        Specifies the minimum quantization index to be used. Must be 0 for
        lossless codecs.
    minimum_slice_size_scaler : int
        Specifies the minimum slice_size_scaler to be used for high quality
        pictures. Ignored in low delay mode.

    Returns
    =======
    data_units : [:py:class:`vc2_conformance.bitstream.DataUnit`, ...]
    """
    if codec_features["fragment_slice_count"] == 0:
        return [
            make_picture_parse_data_unit(
                codec_features, picture, minimum_qindex, minimum_slice_size_scaler
            )
        ]
    else:
        return make_fragment_parse_data_units(
            codec_features, picture, minimum_qindex, minimum_slice_size_scaler
        )



r"""
The :py:mod:`vc2_conformance.bitstream` module implements facilities for
deserialising, displaying, manipulating and serialising VC-2 bitstreams,
including non-conformant streams, at a low-level.

This documentation begins with an overview of how bitstreams can be serialised,
deserialised and represented as Python data structures using this module. This
is followed by an in-depth description of how the serialiser and deserialisers
work internally.

How the serialiser/deserialiser module is used
----------------------------------------------

This module is used by various parts of the VC-2 conformance software, for
example:

* The :ref:`vc2-bitstream-viewer` utility uses this module to produce human
  readable, hierarchical descriptions of bitstreams.
* The test case generators in :py:mod:`vc2_conformance.test_cases` use this
  module to manipulate bitstreams, for example by tweaking values or filling
  padding bits with specific data.
* The VC-2 encoder in :py:mod:`vc2_conformance.encoder` produces deserialised
  bitstreams directly for later serialisation by this module.
* The conformance software's own test suite makes extensive use of this module.

.. note::

    This module is *not* used by the bitstream validator
    (:py:mod:`vc2_conformance.decoder`) which instead operates directly on the
    binary bitstream instead.

This module consists of two main parts: the
:py:mod:`~vc2_conformance.bitstream.serdes` framework for building serialisers
and deserialisers and a serialiser/deserialiser for VC-2.

The :py:mod:`~vc2_conformance.bitstream.serdes` framework allows serialisers
and deserialisers to be constructed *directly* from the VC-2 pseudocode,
ensuring a high chance of correctness.

The VC-2 serialiser/deserialiser, implemented using the
:py:mod:`~vc2_conformance.bitstream.serdes` framework defines a series of
Python data structures which may be used to describe a VC-2 bitstream.


Quick-start guide
-----------------

Before diving into the details, we'll briefly give few quick examples which
illustrate how this module is used below. We'll show how a small bitstream can
be explicitly described as a Python data structure, serialised and then
deserialised again. We'll skim over the details (and ignore a number of
important but minor features) in the process.


Deserialised bitstream data structures
``````````````````````````````````````

A VC-2 bitstream can be described hierarchically as a series of dictionaries
and lists. For example, the following structure describes a minimal VC-2
bitstream containing just a single end-of-sequence data unit::

    >>> from bitarray import bitarray

    >>> # A minimal bitstream...
    >>> bitstream_description = {
    ...     # ...consisting of a single sequence...
    ...     "sequences": [
    ...         {
    ...             # ...with a single data unit...
    ...             "data_units": [
    ...                 {
    ...                     # ...which is an end-of-sequence data unit
    ...                     "parse_info": {
    ...                         "padding": bitarray(),  # No byte alignment bits
    ...                         "parse_info_prefix": 0x42424344,
    ...                         "parse_code": 0x10,  # End of sequence
    ...                         "next_parse_offset": 0,
    ...                         "previous_parse_offset": 0,
    ...                     },
    ...                 },
    ...             ],
    ...         },
    ...     ],
    ... }

To make this somewhat clearer and more robust, a set of :py:mod:`fixeddicts
<vc2_conformance.fixeddict>` are provided which may be used instead of bare
Python dictionaries. The :py:mod:`vc2_data_tables` package also includes many
helpful constant definitions.  Together, these make it easier to see what's
going on while also eliminating simple mistakes like misspelling a field name.

.. note::

    :py:mod:`fixeddicts <vc2_conformance.fixeddict>` are subclasses of Python's
    native :py:class:`dict` type with the following extra features:

    * They allow only a permitted set of key names to be used.
    * The ``__str__`` implementation produces an easier to read pretty-printed
      format.

Using these types, our example now looks like::

    >>> from vc2_data_tables import PARSE_INFO_PREFIX, ParseCodes
    >>> from vc2_conformance.bitstream import Stream, Sequence, DataUnit, ParseInfo

    >>> bitstream_description = Stream(
    ...     sequences=[
    ...         Sequence(
    ...             data_units=[
    ...                 DataUnit(
    ...                     parse_info=ParseInfo(
    ...                         padding=bitarray(),  # No byte alignment bits
    ...                         parse_info_prefix=PARSE_INFO_PREFIX,
    ...                         parse_code=ParseCodes.end_of_sequence,
    ...                         next_parse_offset=0,
    ...                         previous_parse_offset=0,
    ...                     ),
    ...                 ),
    ...             ],
    ...         ),
    ...     ],
    ... )

See :ref:`bitstream-fixeddicts` for details of the expected hierarchy of a
deserialised bitstream (and the :py:mod:`~vc2_conformance.fixeddict` dictionary
types provided).


Serialising bitstreams
``````````````````````

To serialise our bitstream into binary form, we can use the following (which
we'll unpick afterwards)::

    >>> from vc2_conformance.bitstream import BitstreamWriter, Serialiser, parse_stream
    >>> from vc2_conformance.pseudocode import State

    >>> with open("/path/to/bitstream.vc2", "wb") as f:
    ...     with Serialiser(BitstreamWriter(f), bitstream_description) as ser:
    ...         parse_stream(ser, State())

In the example above, ``parse_stream`` is (a special version of) the
``parse_stream`` VC-2 pseudocode function provided by the
:py:mod:`vc2_conformance.bitstream` module. This pseudocode function would
normally decode a VC-2 stream (as per the VC-2 specification), however this
modified version may be used to serialise (or deserialise) a bitstream.  The
modified function takes an extra first argument, a
:py:class:`~vc2_conformance.bitstream.serdes.Serialiser` in this case, which it
will use to produce the serialised bitstream.

The :py:class:`~vc2_conformance.bitstream.serdes.Serialiser` class takes two
arguments in this example: a
:py:class:`~vc2_conformance.bitstream.io.BitstreamWriter` and the data
structure to serialise (``bitstream_description`` in our example).

The :py:class:`~vc2_conformance.bitstream.io.BitstreamWriter` is a wrapper for
file-like objects which provides additional bitwise I/O operations used during
serialisation.

The second argument to :py:class:`~vc2_conformance.bitstream.serdes.Serialiser`
is the deserialised data structure to be serialised. This may be an ordinary
Python :py:class:`dicts <dict>` or a :py:mod:`~vc2_conformance.fixeddict`.

.. note::

    It is possible to serialise (or deserialise) components of a bitstream in
    isolation by using other pseudocode functions in place of ``parse_stream``.
    In this case, the data structure provided to the
    :py:class:`~vc2_conformance.bitstream.serdes.Serialiser` must match the
    shape expected by the modified pseudocode function used. See
    :ref:`bitstream-fixeddicts` for an enumeration of the pseudocode functions
    available and the expected data structure.


Autofilling bitstream values
````````````````````````````

In the example above we explicitly spelt out every field in the bitstream --
including the empty padding field!  If we had omitted this field, the
serialiser will produce an error because it wouldn't know what padding bits we
wanted it to use in the stream. However, often we don't care about details such
as these and so the :py:class:`~vc2_conformance.bitstream.Serialiser` can
optionally 'autofill' certain values which weren't given in the deserialised
data structure.

The :py:class:`~vc2_conformance.bitstream.Serialiser` class takes an optional
third argument which it uses to autofill missing values. A sensible set of a
autofill values is provided in
:py:mod:`vc2_conformance.bitstream.vc2_default_values` allowing us to rewrite
our example like so::

    >>> from vc2_conformance.bitstream import vc2_default_values

    >>> concise_bitstream_description = Stream(
    ...     sequences=[
    ...         Sequence(
    ...             data_units=[
    ...                 DataUnit(
    ...                     parse_info=ParseInfo(
    ...                         parse_code=ParseCodes.end_of_sequence,
    ...                         next_parse_offset=0,
    ...                         previous_parse_offset=0,
    ...                     ),
    ...                 ),
    ...             ],
    ...         ),
    ...     ],
    ... )

    >>> with open("/path/to/bitstream.vc2", "wb") as f:
    ...     with Serialiser(
    ...         BitstreamWriter(f),
    ...         concise_bitstream_description,
    ...         vc2_default_values,
    ...     ) as ser:
    ...         parse_stream(ser, State())


This time we were able to omit the byte-alignment padding value and parse info
prefix which the serialiser autofilled with zeros and 0x42424344 respectively.

The default values for all fields are given in :ref:`bitstream-fixeddicts`.

Unfortunately, when using the mechanism above, autofill values are not provided
for every field in a bitstream (or for fields listed as autofilled with
``<AUTO>`` in the documentation). For instance, picture numbers and parse
offset values must still be calculated and specified explicitly. For a more
complete bitstream autofill solution the
:py:func:`vc2_conformance.bitstream.autofill_and_serialise_stream` utility
function is provided.

The :py:func:`~vc2_conformance.bitstream.autofill_and_serialise_stream`
function can autofill most values including picture numbers, and parse offset
fields (which marked as ``<AUTO>`` in the documentation). It also provides a
more concise wrapper around the serialisation process.

Using :py:func:`~vc2_conformance.bitstream.autofill_and_serialise_stream` our
example now becomes::

    >>> from vc2_conformance.bitstream import autofill_and_serialise_stream

    >>> very_concise_bitstream_description = Stream(
    ...     sequences=[
    ...         Sequence(
    ...             data_units=[
    ...                 DataUnit(
    ...                     parse_info=ParseInfo(
    ...                         parse_code=ParseCodes.end_of_sequence,
    ...                     ),
    ...                 ),
    ...             ],
    ...         ),
    ...     ],
    ... )

    >>> with open("/path/to/bitstream.vc2", "wb") as f:
    ...     autofill_and_serialise_stream(f, very_concise_bitstream_description)

Notice that this time we could omit all but the ``parse_code`` field.

.. note::

    The :py:func:`~vc2_conformance.bitstream.autofill_and_serialise_stream`
    function only supports serialisation of entire :py:class:`Streams
    <vc2_conformance.bitstream.Stream>` and cannot be used to serialise smaller
    pieces of a bitstream in isolation.


Deserialising bitstreams
````````````````````````

To deserialise a bitstream again, the process is similar::

    >>> from vc2_conformance.bitstream import BitstreamReader, Deserialiser

    >>> with open("/tmp/bitstream.vc2", "rb") as f:
    ...     with Deserialiser(BitstreamReader(f)) as des:
    ...         parse_stream(des, State())
    >>> deserialised_bitstream = des.context

This time, we pass a :py:class:`~vc2_conformance.bitstream.serdes.Deserialiser`
(which takes a :py:class:`~vc2_conformance.bitstream.io.BitstreamReader` as
argument) into ``parse_stream``. The deserialised bitstream is placed into
``des.``\ :py:attr:`~vc2_conformance.bitstream.serdes.Deserialiser.context` as a hierarchy
of :py:mod:`~vc2_conformance.fixeddicts`. We can then print or interact with the
deserialised data structure just like any other Python object::

    >>> # NB: Fixeddicts produce pretty-printed output when printed!
    >>> print(deserialised_bitstream)
    Stream:
      sequences:
        0: Sequence:
          data_units:
            0: DataUnit:
              parse_info: ParseInfo:
                padding: 0b
                parse_info_prefix: Correct (0x42424344)
                parse_code: end_of_sequence (0x10)
                next_parse_offset: 0
                previous_parse_offset: 0

    >>> data_unit = deserialised_bitstream["sequences"][0]["data_units"][0]
    >>> print(data_unit["parse_info"]["parse_code"])
    16


.. _bitstream-fixeddicts:

Deserialised VC-2 bitstream data types
--------------------------------------

Deserialised VC-2 bitstreams are described by a hierarchy of
:py:class:`fixeddicts <vc2_conformance.fixeddict>`, exported in
:py:mod:`vc2_conformance.bitstream`.  Each
:py:class:`~vc2_conformance.fixeddict` represents the data read by a particular
VC-2 pseudocode function. Special implementations of these functions are
provided in :py:mod:`vc2_conformance.bitstream` which may be used to serialise
and deserialise VC-2 bitstreams (or individual parts thereof).

..
    The following RST directives are implemented by
    ``docs/source/_ext/bitstream_fixeddicts.py``. These one-off directives
    generate a complete listing of:

    * The fixeddict types defined in
      :py:mod:`vc2_conformance.bitstream.vc2_fixeddicts`
    * The pseudocode functions defined in :py:mod:`vc2_conformance.bitstream.vc2`
    * The autofill values defined in
      :py:mod:`vc2_conformance.bitstream.vc2_autofill`

.. autobitstreamfixeddictstable::

.. autobitstreamfixeddicts::

:py:mod:`~vc2_conformance.bitstream.serdes`: A serialiser/deserialiser framework
--------------------------------------------------------------------------------

.. automodule:: vc2_conformance.bitstream.serdes


Low-level bitstream IO
----------------------

.. automodule:: vc2_conformance.bitstream.io

Fixeddicts and pseudocode
-------------------------

.. automodule:: vc2_conformance.bitstream.vc2_fixeddicts

.. automodule:: vc2_conformance.bitstream.vc2


Autofill
--------

.. automodule:: vc2_conformance.bitstream.vc2_autofill


Metadata
--------

.. automodule:: vc2_conformance.bitstream.metadata

"""

from vc2_conformance.bitstream.exceptions import *

# Exp-golomb code length calculators
from vc2_conformance.bitstream.exp_golomb import *

# Low-level bitwise file reading/writing
from vc2_conformance.bitstream.io import *

# Generic bitstream serialisation/deserialisation framework
from vc2_conformance.bitstream.serdes import *

# VC-2 specific parts
from vc2_conformance.bitstream.vc2 import *
from vc2_conformance.bitstream.vc2_fixeddicts import *
from vc2_conformance.bitstream.vc2_autofill import *

# Metadata for introspection purposes
from vc2_conformance.bitstream.metadata import *



"""
:py:mod:`vc2_conformance.bitstream.exp_golomb`: Exp-golomb code length calculators
==================================================================================
"""

from vc2_conformance.bitstream.exceptions import OutOfRangeError

__all__ = [
    "exp_golomb_length",
    "signed_exp_golomb_length",
]


def exp_golomb_length(value):
    """
    Return the length (in bits) of the unsigned exp-golomb representation of
    value.

    An :py:exc:`~.OutOfRangeError` will be raised if
    a negative value is provided.
    """
    if value < 0:
        raise OutOfRangeError(value)

    return (((value + 1).bit_length() - 1) * 2) + 1


def signed_exp_golomb_length(value):
    """
    Return the length (in bits) of the signed exp-golomb representation of
    value.
    """
    length = exp_golomb_length(abs(value))
    if value != 0:
        length += 1
    return length



"""
Tests which verify that codecs correctly process streams formed of multiple
concatenated sequences (10.3).
"""

from copy import deepcopy

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import decoder_test_case_generator

from vc2_conformance.picture_generators import mid_gray

from vc2_conformance.encoder import make_sequence


@decoder_test_case_generator
def concatenated_sequences(codec_features):
    """
    **Tests that streams containing multiple concatenated sequences can be
    decoded.**

    A stream consisting of the concatenation of two sequences (10.3) with one
    frame each, the first picture is given picture number zero in both
    sequences.
    """
    sequence = make_sequence(
        codec_features,
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
    )

    return Stream(sequences=[sequence, deepcopy(sequence)])



r"""
The :py:mod:`vc2_conformance.fixeddict` module provides the
:py:func:`fixeddict` function for creating new :py:class:`dict` subclasses
which permit only certain keys to be used.  These new types may be used like
ordinary Python dictionaries but add three main features:

* Explicitness -- The VC-2 pseudocode creates and uses many dictionaries
  (called 'maps' in the specification) in place of struct-like objects.
  :py:mod:`fixeddicts <vc2_conformance.fixeddict>` provide a way to give these
  clear names.
* Avoidance of typos -- Misspelt key names will result in a
  :py:exc:`FixedDictKeyError`.
* Better pretty printing -- See more below...


Tutorial
--------

Using :py:func:`fixeddict`, dictionary-like types with well defined fields can
be described like so::

    >>> from vc2_conformance.fixeddict import fixeddict

    >>> FrameSize = fixeddict(
    ...     "FrameSize",
    ...     "custom_dimensions_flag",
    ...     "frame_width",
    ...     "frame_height",
    ... )

This produces a 'dict' subclass called ``FrameSize`` with all of the usual
dictionary behaviour but which only allows the specified keys to be used::

    >>> f = FrameSize()
    >>> f["custom_dimensions_flag"] = True
    >>> f["frame_width"] = 1920
    >>> f["frame_height"] = 1080

    >>> f["frame_width"]
    1920

    >>> f["not_in_fixeddict"] = 123
    Traceback (most recent call last):
      ...
    FixedDictKeyError: 'not_in_fixeddict'


To improve readability when producing string representations of VC-2 data
structures, the generated dictionary types have a 'pretty' string
representation.

    >>> print(f)
    FrameSize:
      custom_dimensions_flag: True
      frame_width: 1920
      frame_height: 1080

To further improve the readability of this output, custom string formatting
functions may be provided for each entry in the dictionary. To define these,
:py:class:`Entry` instances must be used in place of key name strings like
so::

    >>> from vc2_conformance.string_formatters import Hex
    >>> from vc2_data_tables import ParseCodes  # An IntEnum
    >>> ParseInfo = fixeddict(
    ...     "ParseInfo",
    ...     Entry("parse_info_prefix", formatter=Hex(8)),
    ...     Entry("parse_code", enum=ParseCodes, formatter=Hex(2)),
    ...     Entry("next_parse_offset"),
    ...     Entry("previous_parse_offset"),

    >>> pi = ParseInfo(
    ...     parse_info_prefix=0x42424344,
    ...     parse_code=0x10,
    ...     next_parse_offset=0,
    ...     previous_parse_offset=0,
    ... )
    >>> str(pi)
    ParseInfo:
       parse_info_prefix: 0x42424344
       parse_code: end_of_sequence (0x10)
       next_parse_offset: 0
       previous_parse_offset: 0

See the :py:mod:`vc2_conformance.string_formatters` module for a set of useful
string formatting functions.

Finally, documentation can optionally be added in the form of ``help`` and
``help_type`` arguments which will combined into the generated type's
docstring::

    >>> ParseInfo = fixeddict(
    ...     "ParseInfo",
    ...     Entry("parse_info_prefix", formatter=Hex(8), help_type="int", help="Always 0x42424344"),
    ...     Entry("parse_code", enum=ParseCodes, formatter=Hex(2), help_type="int"),
    ...     Entry("next_parse_offset", help_type="int"),
    ...     Entry("previous_parse_offset", help_type="int"),
    ...     help="A deserialised parse info block.",
    ... )

    >>> print(ParseInfo.__doc__)
    A deserialised parse info block.
    <BLANKLINE>
    Parameters
    ==========
    parse_info_prefix : int
        Always 0x42424344
    parse_code : int
    next_parse_offset : int
    previous_parse_offset : int


API
---

.. autofunction:: fixeddict

.. autoclass:: Entry

.. autoexception:: FixedDictKeyError

"""

import sys

from collections import OrderedDict

from textwrap import dedent

from vc2_conformance.string_utils import indent


__all__ = [
    "fixeddict",
    "Entry",
    "FixedDictKeyError",
]


class Entry(object):
    """
    Defines advanced properties of of an entry in a :py:func:`fixeddict`
    dictionary.

    All constructor arguments, except name, are keyword-only.

    Parameters
    ==========
    name : str
        The name of this entry in the dictionary.
    formatter : function(value) -> string
        A function which takes a value and returns a string representation
        to use when printing this value as a string. Defaults to 'str'.
    friendly_formatter : function(value) -> string
        If provided, when converting this value to a string, this function
        will be used to generate a 'friendly' name for this value. This
        will be followed by the actual value in brackets. If this function
        returns None, only the actual value will be shown (without
        brackets).
    enum : :py:class:`~enum.Enum`
        A convenience interface which is equivalent to the following
        ``formatter`` argument::

            def enum_formatter(value):
                try:
                    return str(MyEnum(value).value)
                except ValueError:
                    return str(value)

        And the following ``friendly_formatter`` argument::

            def friendly_enum_formatter(value):
                try:
                    return MyEnum(value).name
                except ValueError:
                    return None

        If ``formatter`` or ``friendly_formatter`` are provided in addition
        to ``enum``, they will override the functions implicitly defined by
        ``enum``.
    help : str
        Optional documentation string.
    help_type : str
        Optional string describing the type of the entry.
    """

    def __init__(self, name, **kwargs):
        self.name = name

        if "enum" in kwargs:
            enum_type = kwargs.pop("enum")

            def enum_formatter(value):
                try:
                    return str(enum_type(value).value)
                except ValueError:
                    return str(value)

            kwargs.setdefault("formatter", enum_formatter)

            def friendly_enum_formatter(value):
                try:
                    return enum_type(value).name
                except ValueError:
                    return None

            kwargs.setdefault("friendly_formatter", friendly_enum_formatter)

        self.formatter = kwargs.pop("formatter", str)
        self.friendly_formatter = kwargs.pop("friendly_formatter", None)

        self.help = kwargs.pop("help", None)
        if self.help is not None:
            self.help = dedent(self.help).strip()

        self.help_type = kwargs.pop("help_type", None)
        if self.help_type is not None:
            self.help_type = dedent(self.help_type).strip()

        if kwargs:
            raise TypeError(
                "unexpected keyword arguments: {} for {}".format(
                    ", ".join(kwargs.keys()), self.__class__.__name__
                )
            )

    def to_string(self, value):
        """
        Convert a value to a string according to the specification in this
        :py:class:`Entry`.
        """
        value_string = self.formatter(value)

        if self.friendly_formatter is not None:
            friendly_string = self.friendly_formatter(value)
            if friendly_string is not None:
                value_string = "{} ({})".format(friendly_string, value_string)

        return value_string


class FixedDictKeyError(KeyError):
    """
    A :py:exc:`KeyError` which also includes information about which fixeddict
    dictionary it was produced by.

    Attributes
    ==========
    key
        The key which was accessed.
    fixeddict_class
        The :py:mod:`~vc2_conformance.fixeddict` type of the dictionary used.
    """

    def __init__(self, key, fixeddict_class):
        super(FixedDictKeyError, self).__init__(key)
        self.key = key
        self.fixeddict_class = fixeddict_class

    def __str__(self):
        return "{!r} not allowed in {}".format(self.key, self.fixeddict_class.__name__)


def fixeddict(name, *entries, **kwargs):
    """
    Create a fixed-entry dictionary.

    A fixed-entry dictionary is a :py:class:`dict` subclass which permits only
    a preset list of key names.

    The first argument is the name of the created class, the remaining
    arguments may be strings or :py:class:`Entry` instances describing the
    allowed entries in the dictionary.

    Example usage:

        >>> ExampleDict = fixeddict(
        ...     "ExampleDict",
        ...     "attr",
        ...     Entry("attr_with_default"),
        ... )

    Instances of the dictionary can be created like an ordinary dictionary::

        >>> d = ExampleDict(attr=10, attr_with_default=20)
        >>> d["attr"]
        10
        >>> d["attr_with_default"]
        20

    The string format of generated dictionaries includes certain
    pretty-printing behaviour (see :py:class:`Entry`) and will also omit any
    entries whose name is prefixed with an underscore (``_``).

    The class itself will have a static (and read-only) attribute
    ``entry_objs`` which is a :py;class:`collections.OrderedDict` mapping from
    entry name to :py:class:`Entry` object in the dictionary.

    The keyword-only argument, 'module' may be provided which overrides the
    ``__module__`` value of the returned fixeddict type. (By default the module
    name is inferred using runtime stack inspection, if possible). This must be
    set correctly for this type to be picklable.

    The keyword-only argument 'help' may be used to set the docstring of the
    returned class. This will automatically be appended with the list of
    entries allowed (and their help strings).
    """
    # Extract keyword-only arguments
    module = kwargs.pop("module", None)
    help = kwargs.pop("help", None)
    assert not kwargs, "Got unexpected keyword arguments: {}".format(", ".join(kwargs))

    if help is not None:
        help = dedent(help).strip()

    # Collect the list of Entry instances defined for this class
    # {name: Entry, ...}
    entry_objs = OrderedDict(
        (entry.name, entry)
        for entry in (arg if isinstance(arg, Entry) else Entry(arg) for arg in entries)
    )

    # Create all of the methods which will be added to the class
    __dict__ = {}

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # Check for invalid names
        for name in self.keys():
            if name not in entry_objs:
                raise FixedDictKeyError(name, self.__class__)

    __dict__["__init__"] = __init__

    __dict__["help"] = help
    __dict__["__doc__"] = "{}\n\nParameters\n==========\n{}\n".format(
        help if help is not None else "A :py:mod:`~vc2_conformance.fixeddict`.",
        "\n".join(
            "{}{}{}".format(
                entry.name,
                (" : " + entry.help_type if entry.help_type is not None else ""),
                (("\n" + indent(entry.help, "    ")) if entry.help is not None else ""),
            )
            for entry in entry_objs.values()
        ),
    )

    def __setitem__(self, key, value):
        if key in entry_objs:
            return dict.__setitem__(self, key, value)
        else:
            raise FixedDictKeyError(key, self.__class__)

    __dict__["__setitem__"] = __setitem__

    def setdefault(self, key, value):
        if key in entry_objs:
            return dict.setdefault(self, key, value)
        else:
            raise FixedDictKeyError(key, self.__class__)

    __dict__["setdefault"] = setdefault

    def update(self, E=None, **F):
        # Using the naming convention defined by 'help(dict.setdefault)'
        if E is not None:
            if hasattr(E, "keys"):
                for k in E:
                    self[k] = E[k]
            else:
                for k, v in E:
                    self[k] = v
        for k in F:
            self[k] = F[k]

    __dict__["update"] = update

    def __repr__(self):
        return "{}({{{}}})".format(
            self.__class__.__name__,
            ", ".join(
                "{!r}: {!r}".format(name, self[name])
                for name, entry_obj in entry_objs.items()
                if name in self
            ),
        )

    __dict__["__repr__"] = __repr__

    def __str__(self):
        if len(self) == 0:
            return self.__class__.__name__
        else:
            return "{}:\n{}".format(
                self.__class__.__name__,
                "\n".join(
                    indent("{}: {}".format(name, entry_obj.to_string(self[name])))
                    for name, entry_obj in entry_objs.items()
                    if name in self and not name.startswith("_")
                ),
            )

    __dict__["__str__"] = __str__

    def copy(self):
        return self.__class__(self)

    __dict__["copy"] = copy

    # Support pickling/unpickling (part 1).
    #
    # Dictionaries have their own magic behaviour by default under pickle so we
    # must explicitly tell pickle how to handle this type.
    def __getstate__(self):
        return dict(self)

    __dict__["__getstate__"] = __getstate__

    def __setstate__(self, state):
        self.update(state)

    __dict__["__setstate__"] = __setstate__

    def __reduce__(self):
        return (
            type(self),
            (),
            self.__getstate__(),
        )

    __dict__["__reduce__"] = __reduce__

    cls = type(name, (dict,), __dict__)

    setattr(cls, "entry_objs", entry_objs)

    # Support pickling/unpickling (part 2)
    #
    # Setting the __module__ class attributes tells pickle where to find this
    # type when unpickling.
    if module is None:
        # Detect the module of the caller by inspecting the stack. This is a
        # bit gross, and won't work under all Python interpreters, but is what
        # enum.Enum also has to do and if its good enough for the stdlib, its
        # good enough for this...
        try:
            module = sys._getframe(1).f_globals["__name__"]
        except (AttributeError, ValueError, KeyError):
            pass
    if module is not None:
        setattr(cls, "__module__", module)

    return cls



"""
The :py:mod:`vc2_conformance.decoder.stream` module contains pseudocode functions from
(10) VC-2 Stream Syntax.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.pseudocode.parse_code_functions import (
    is_seq_header,
    is_picture,
    is_fragment,
    is_auxiliary_data,
    is_padding_data,
    is_end_of_sequence,
)

from vc2_data_tables import (
    PARSE_INFO_PREFIX,
    PARSE_INFO_HEADER_BYTES,
    ParseCodes,
    PROFILES,
    PictureCodingModes,
)

from vc2_conformance.version_constraints import (
    MINIMUM_MAJOR_VERSION,
    parse_code_version_implication,
)

from vc2_conformance.decoder.exceptions import (
    BadParseInfoPrefix,
    BadParseCode,
    InconsistentNextParseOffset,
    MissingNextParseOffset,
    InvalidNextParseOffset,
    NonZeroNextParseOffsetAtEndOfSequence,
    InconsistentPreviousParseOffset,
    NonZeroPreviousParseOffsetAtStartOfSequence,
    GenericInvalidSequence,
    LevelInvalidSequence,
    ParseCodeNotAllowedInProfile,
    OddNumberOfFieldsInSequence,
    SequenceContainsIncompleteFragmentedPicture,
    PictureInterleavedWithFragmentedPicture,
    ParseCodeNotSupportedByVersion,
)

from vc2_conformance.symbol_re import Matcher

from vc2_conformance.pseudocode.picture_decoding import picture_decode

from vc2_conformance.decoder.assertions import (
    assert_in_enum,
    assert_parse_code_in_sequence,
    assert_parse_code_sequence_ended,
    log_version_lower_bound,
    assert_major_version_is_minimal,
)

from vc2_conformance.pseudocode.state import reset_state

from vc2_conformance.decoder.io import (
    is_end_of_stream,
    tell,
    byte_align,
    read_uint_lit,
)

from vc2_conformance.decoder.sequence_header import sequence_header
from vc2_conformance.decoder.picture_syntax import picture_parse
from vc2_conformance.decoder.fragment_syntax import fragment_parse


__all__ = [
    "parse_stream",
    "parse_sequence",
    "auxiliary_data",
    "padding",
    "parse_info",
]


@ref_pseudocode
def parse_stream(state):
    """
    (10.3) Parse a complete VC-2 stream (consisting of several concatenated
    sequences).
    """
    while not is_end_of_stream(state):
        parse_sequence(state)


@ref_pseudocode
def parse_sequence(state):
    """
    (10.4.1) Parse a complete VC-2 sequence.
    """
    reset_state(state)

    # (10.4.1) Check that the sequence starts with a sequence_header and ends
    # with and end_of_sequence.
    ## Begin not in spec
    state["_generic_sequence_matcher"] = Matcher("sequence_header .* end_of_sequence")
    ## End not in spec

    state["_num_pictures_in_sequence"] = 0  ## Not in spec
    state["_fragment_slices_remaining"] = 0  ## Not in spec

    parse_info(state)
    while not is_end_of_sequence(state):
        if is_seq_header(state):
            state["video_parameters"] = sequence_header(state)
        elif is_picture(state):
            # Errata: fragments under-specified in spec
            #
            # (14.2) Picture data units may not be interleaved with in-progress
            # fragmented pictures.
            ## Begin not in spec
            if state["_fragment_slices_remaining"] != 0:
                raise PictureInterleavedWithFragmentedPicture(
                    state["_picture_initial_fragment_offset"],
                    tell(state),
                    state["fragment_slices_received"],
                    state["_fragment_slices_remaining"],
                )
            ## End not in spec

            picture_parse(state)
            picture_decode(state)
        elif is_fragment(state):
            fragment_parse(state)
            if state["fragmented_picture_done"]:
                picture_decode(state)
        elif is_auxiliary_data(state):
            auxiliary_data(state)
        elif is_padding_data(state):
            padding(state)
        parse_info(state)

    # (10.4.1) and (C.3) Check sequence structure allowed to end at this point
    ## Begin not in spec
    assert_parse_code_sequence_ended(
        state["_generic_sequence_matcher"], GenericInvalidSequence
    )
    if "_level_sequence_matcher" in state:
        assert_parse_code_sequence_ended(
            state["_level_sequence_matcher"],
            LevelInvalidSequence,
            state["level"],
        )
    ## End not in spec

    # Errata: fragments under-specified in spec
    #
    # (14.2) Ensure that any fragmented picture has been received completely by
    # the end of the stream.
    ## Begin not in spec
    if state["_fragment_slices_remaining"] != 0:
        raise SequenceContainsIncompleteFragmentedPicture(
            state["_picture_initial_fragment_offset"],
            state["fragment_slices_received"],
            state["_fragment_slices_remaining"],
        )
    ## End not in spec

    # (10.4.3) When pictures are fields, a sequence should contain a whole
    # number of frames
    ## Begin not in spec
    if state["picture_coding_mode"] == PictureCodingModes.pictures_are_fields:
        if state["_num_pictures_in_sequence"] % 2 != 0:
            raise OddNumberOfFieldsInSequence(state["_num_pictures_in_sequence"])
    ## End not in spec

    # (11.2.2) Check that the major_version number used was the lowest which
    # would be permissible for the features actually used
    assert_major_version_is_minimal(state)  ## Not in spec


@ref_pseudocode
def auxiliary_data(state):
    """(10.4.4)"""
    for i in range(1, state["next_parse_offset"] - 12):
        read_uint_lit(state, 1)


@ref_pseudocode
def padding(state):
    """(10.4.5)"""
    for i in range(1, state["next_parse_offset"] - 12):
        read_uint_lit(state, 1)


@ref_pseudocode
def parse_info(state):
    """(10.5.1)"""
    byte_align(state)

    # (10.5.1) Check that the previous parse_info's next_parse_offset was
    # calculated correctly
    ## Begin not in spec
    this_parse_info_offset = tell(state)[0]
    last_parse_info_offset = state.get("_last_parse_info_offset", None)

    if state.get("next_parse_offset"):
        true_parse_offset = this_parse_info_offset - last_parse_info_offset
        last_next_parse_offset = state["next_parse_offset"]
        if last_next_parse_offset != 0 and last_next_parse_offset != true_parse_offset:
            raise InconsistentNextParseOffset(
                last_parse_info_offset,
                last_next_parse_offset,
                true_parse_offset,
            )
    ## End not in spec

    # (10.5.1) Capture and check the parse_info prefix
    ### read_uint_lit(state, 4)
    ## Begin not in spec
    prefix = read_uint_lit(state, 4)
    if prefix != PARSE_INFO_PREFIX:
        raise BadParseInfoPrefix(prefix)
    ## End not in spec

    # (10.5.1) Check the parse_code is a supported value
    state["parse_code"] = read_uint_lit(state, 1)
    assert_in_enum(state["parse_code"], ParseCodes, BadParseCode)  ## Not in spec

    # (10.4.1) Check that the sequence starts with a sequence_header and ends
    # with and end_of_sequence.
    ## Begin not in spec
    assert_parse_code_in_sequence(
        state["parse_code"],
        state["_generic_sequence_matcher"],
        GenericInvalidSequence,
    )
    ## End not in spec

    # (C.3) Check that the sequence follows the pattern dictated by the current
    # level (NB: this matcher is populated later by parse_parameters (11.2.1)
    # when the level is read for the first time.
    ## Begin not in spec
    if "_level_sequence_matcher" in state:
        assert_parse_code_in_sequence(
            state["parse_code"],
            state["_level_sequence_matcher"],
            LevelInvalidSequence,
            state["level"],
        )
    ## End not in spec

    # (C.2.2) Ensure that only the profile-permitted parse codes are used
    ## Begin not in spec
    if "profile" in state:
        profile_params = PROFILES[state["profile"]]
        if state["parse_code"] not in profile_params.allowed_parse_codes:
            raise ParseCodeNotAllowedInProfile(state["parse_code"], state["profile"])
    ## End not in spec

    # (11.2.2) Check that the parse code used is supported by the current
    # major_version.
    ## Begin not in spec
    minimum_required_version = parse_code_version_implication(state["parse_code"])
    # NB: If this is the sequence header at the start of a sequence we may not
    # know the major_version yet. In that case, we can safely skip this test as
    # the sequence header is supported by all versions.
    major_version = state.get("major_version", MINIMUM_MAJOR_VERSION)
    if major_version < minimum_required_version:
        raise ParseCodeNotSupportedByVersion(state["parse_code"], major_version)
    log_version_lower_bound(state, minimum_required_version)
    ## End not in spec

    # (10.5.1) Check that the next_parse_offset holds a plausible value
    state["next_parse_offset"] = read_uint_lit(state, 4)
    ## Begin not in spec
    if state["parse_code"] == ParseCodes.end_of_sequence:
        # (10.5.1) End of stream must have '0' as next offset
        if state["next_parse_offset"] != 0:
            raise NonZeroNextParseOffsetAtEndOfSequence(state["next_parse_offset"])
    elif not (is_picture(state) or is_fragment(state)):
        # (10.5.1) Non-picture containing blocks *must* have a non-zero offset
        if state["next_parse_offset"] == 0:
            raise MissingNextParseOffset(state["parse_code"])

    # (10.5.1) Offsets pointing inside this parse_info bock are always invalid
    if 1 <= state["next_parse_offset"] < PARSE_INFO_HEADER_BYTES:
        raise InvalidNextParseOffset(state["next_parse_offset"])
    ## End not in spec

    # (10.5.1) Check that the previous parse offset was calculated correctly
    state["previous_parse_offset"] = read_uint_lit(state, 4)
    ## Begin not in spec
    if last_parse_info_offset is None:
        # (10.5.1) This is the first parse_info encountered, must be zero
        if state["previous_parse_offset"] != 0:
            raise NonZeroPreviousParseOffsetAtStartOfSequence(
                state["previous_parse_offset"]
            )
    else:
        # (10.5.1) Previous offset must be present and calculated correctly otherwise
        true_previous_parse_offset = this_parse_info_offset - last_parse_info_offset
        if state["previous_parse_offset"] != true_previous_parse_offset:
            raise InconsistentPreviousParseOffset(
                last_parse_info_offset,
                state["previous_parse_offset"],
                true_parse_offset,
            )
    ## End not in spec

    state["_last_parse_info_offset"] = this_parse_info_offset  ## Not in spec



from vc2_conformance.test_cases import EncoderTestSequence


def picture_generator_to_test_case(picture_generator, codec_features):
    """
    Internal utility. Takes a function from
    :py:mod:`vc2_conformance.picture_generators` and a set of
    :py:class:`~vc2_conformance.codec_features.CodecFeatures` and
    returns a :py:class:`~vc2_conformance.test_cases.EncoderTestSequence`.
    """
    out = EncoderTestSequence(
        pictures=[],
        video_parameters=codec_features["video_parameters"],
        picture_coding_mode=codec_features["picture_coding_mode"],
    )

    for picture in picture_generator(
        codec_features["video_parameters"],
        codec_features["picture_coding_mode"],
    ):
        out.pictures.append(picture)

    return out



"""
The :py:mod:`vc2_conformance.decoder.transform_data_syntax` module contains pseudocode
functions from (13) Transform data syntax.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.pseudocode.vc2_math import (
    intlog2,
    mean,
)

from vc2_conformance.pseudocode.arrays import (
    new_array,
    width,
    height,
)

from vc2_conformance.pseudocode.slice_sizes import (
    subband_width,
    subband_height,
    slice_bytes,
    slice_left,
    slice_right,
    slice_top,
    slice_bottom,
)

from vc2_conformance.pseudocode.parse_code_functions import (
    is_ld,
    is_hq,
    using_dc_prediction,
)

from vc2_conformance.decoder.exceptions import InvalidSliceYLength

from vc2_conformance.decoder.assertions import assert_level_constraint

from vc2_conformance.decoder.io import (
    tell,
    read_uint_lit,
    read_nbits,
    read_sintb,
    flush_inputb,
)

from vc2_conformance.pseudocode.quantization import inverse_quant

__all__ = [
    "initialize_wavelet_data",
    "dc_prediction",
    "transform_data",
    "slice",
    "ld_slice",
    "hq_slice",
    "slice_quantizers",
    "slice_band",
    "color_diff_slice_band",
]


@ref_pseudocode
def dc_prediction(band):
    """(13.4)"""
    for y in range(height(band)):
        for x in range(width(band)):
            if x > 0 and y > 0:
                prediction = mean(band[y][x - 1], band[y - 1][x - 1], band[y - 1][x])
            elif x > 0 and y == 0:
                prediction = band[0][x - 1]
            elif x == 0 and y > 0:
                prediction = band[y - 1][0]
            else:
                prediction = 0
            band[y][x] += prediction


@ref_pseudocode(deviation="inferred_implementation")
def initialize_wavelet_data(state, comp):
    """(13.2.2) Return a ready-to-fill array of transform data arrays."""
    out = {}

    if state["dwt_depth_ho"] == 0:
        out[0] = {
            "LL": new_array(
                subband_height(state, 0, comp), subband_width(state, 0, comp)
            )
        }
    else:
        out[0] = {
            "L": new_array(
                subband_height(state, 0, comp), subband_width(state, 0, comp)
            )
        }
        for level in range(1, state["dwt_depth_ho"] + 1):
            out[level] = {
                "H": new_array(
                    subband_height(state, level, comp),
                    subband_width(state, level, comp),
                )
            }

    for level in range(
        state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
    ):
        out[level] = {
            orient: new_array(
                subband_height(state, level, comp), subband_width(state, level, comp)
            )
            for orient in ["HL", "LH", "HH"]
        }
    return out


@ref_pseudocode
def transform_data(state):
    """(13.5.2)"""
    state["y_transform"] = initialize_wavelet_data(state, "Y")
    state["c1_transform"] = initialize_wavelet_data(state, "C1")
    state["c2_transform"] = initialize_wavelet_data(state, "C2")
    for sy in range(state["slices_y"]):
        for sx in range(state["slices_x"]):
            slice(state, sx, sy)
    if using_dc_prediction(state):
        if state["dwt_depth_ho"] == 0:
            dc_prediction(state["y_transform"][0]["LL"])
            dc_prediction(state["c1_transform"][0]["LL"])
            dc_prediction(state["c2_transform"][0]["LL"])
        else:
            dc_prediction(state["y_transform"][0]["L"])
            dc_prediction(state["c1_transform"][0]["L"])
            dc_prediction(state["c2_transform"][0]["L"])


@ref_pseudocode
def slice(state, sx, sy):
    """(13.5.2)"""
    if is_ld(state):
        ld_slice(state, sx, sy)
    elif is_hq(state):
        hq_slice(state, sx, sy)


@ref_pseudocode
def ld_slice(state, sx, sy):
    """(13.5.3.1)"""
    slice_bits_left = 8 * slice_bytes(state, sx, sy)

    qindex = read_nbits(state, 7)
    slice_bits_left -= 7
    # Errata: none of the levels currently restrict the qindex
    #
    # (C.3) The quantisation index may be restricted by some levels
    assert_level_constraint(state, "qindex", qindex)  ## Not in spec

    slice_quantizers(state, qindex)

    length_bits = intlog2(8 * slice_bytes(state, sx, sy) - 7)
    slice_y_length = read_nbits(state, length_bits)
    slice_bits_left -= length_bits

    # (13.5.3.1) The slice_y_length is restricted to be within the remaining
    # space in this slice
    ## Begin not in spec
    # NB: slice_bits_left = 8*slice_bytes(state, sx, sy) - 7 - length_bits
    if slice_y_length > slice_bits_left:
        raise InvalidSliceYLength(
            slice_y_length,
            slice_bytes(state, sx, sy),
            sx,
            sy,
        )
    ## End not in spec

    state["bits_left"] = slice_y_length
    if state["dwt_depth_ho"] == 0:
        slice_band(state, "y_transform", 0, "LL", sx, sy)
        for level in range(1, state["dwt_depth"] + 1):
            for orient in ["HL", "LH", "HH"]:
                slice_band(state, "y_transform", level, orient, sx, sy)
    else:
        slice_band(state, "y_transform", 0, "L", sx, sy)
        for level in range(1, state["dwt_depth_ho"] + 1):
            slice_band(state, "y_transform", level, "H", sx, sy)
        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            for orient in ["HL", "LH", "HH"]:
                slice_band(state, "y_transform", level, orient, sx, sy)
    flush_inputb(state)

    slice_bits_left -= slice_y_length
    state["bits_left"] = slice_bits_left
    if state["dwt_depth_ho"] == 0:
        color_diff_slice_band(state, 0, "LL", sx, sy)
        for level in range(1, state["dwt_depth"] + 1):
            for orient in ["HL", "LH", "HH"]:
                color_diff_slice_band(state, level, orient, sx, sy)
    else:
        color_diff_slice_band(state, 0, "L", sx, sy)
        for level in range(1, state["dwt_depth_ho"] + 1):
            color_diff_slice_band(state, level, "H", sx, sy)
        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            for orient in ["HL", "LH", "HH"]:
                color_diff_slice_band(state, level, orient, sx, sy)
    flush_inputb(state)


@ref_pseudocode
def hq_slice(state, sx, sy):
    """(13.5.4)"""
    byte_offset_start = tell(state)[0]  ## Not in spec

    read_uint_lit(state, state["slice_prefix_bytes"])

    qindex = read_uint_lit(state, 1)
    # Errata: none of the levels currently restrict the qindex
    #
    # (C.3) The quantisation index may be restricted by some levels
    assert_level_constraint(state, "qindex", qindex)  ## Not in spec

    slice_quantizers(state, qindex)

    for transform in ["y_transform", "c1_transform", "c2_transform"]:
        length = state["slice_size_scaler"] * read_uint_lit(state, 1)
        state["bits_left"] = 8 * length
        if state["dwt_depth_ho"] == 0:
            slice_band(state, transform, 0, "LL", sx, sy)
            for level in range(1, state["dwt_depth"] + 1):
                for orient in ["HL", "LH", "HH"]:
                    slice_band(state, transform, level, orient, sx, sy)
        else:
            slice_band(state, transform, 0, "L", sx, sy)
            for level in range(1, state["dwt_depth_ho"] + 1):
                slice_band(state, transform, level, "H", sx, sy)
            for level in range(
                state["dwt_depth_ho"] + 1,
                state["dwt_depth_ho"] + state["dwt_depth"] + 1,
            ):
                for orient in ["HL", "LH", "HH"]:
                    slice_band(state, transform, level, orient, sx, sy)
        flush_inputb(state)

    # (C.3) Some levels restrict the total size of HQ slices
    ## Begin not in spec
    byte_offset_end = tell(state)[0]
    total_slice_bytes = byte_offset_end - byte_offset_start
    assert_level_constraint(state, "total_slice_bytes", total_slice_bytes)
    ## End not in spec


@ref_pseudocode
def slice_quantizers(state, qindex):
    """(13.5.5)"""
    # NB: For historical reasons, we use a dict not an array in this
    # implementation.
    ### state["quantizer"] = new_array(state["dwt_depth_ho"] + state["dwt_depth"] + 1)
    state["quantizer"] = {}  ## Not in spec
    if state["dwt_depth_ho"] == 0:
        state["quantizer"][0] = {}
        state["quantizer"][0]["LL"] = max(qindex - state["quant_matrix"][0]["LL"], 0)
        for level in range(1, state["dwt_depth"] + 1):
            state["quantizer"][level] = {}
            for orient in ["HL", "LH", "HH"]:
                qval = max(qindex - state["quant_matrix"][level][orient], 0)
                state["quantizer"][level][orient] = qval
    else:
        state["quantizer"][0] = {}
        state["quantizer"][0]["L"] = max(qindex - state["quant_matrix"][0]["L"], 0)
        for level in range(1, state["dwt_depth_ho"] + 1):
            state["quantizer"][level] = {}
            qval = max(qindex - state["quant_matrix"][level]["H"], 0)
            state["quantizer"][level]["H"] = qval
        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            state["quantizer"][level] = {}
            for orient in ["HL", "LH", "HH"]:
                qval = max(qindex - state["quant_matrix"][level][orient], 0)
                state["quantizer"][level][orient] = qval


@ref_pseudocode
def slice_band(state, transform, level, orient, sx, sy):
    """(13.5.6.3)"""
    if transform == "y_transform":
        comp = "Y"
    elif transform == "c1_transform":
        comp = "C1"
    elif transform == "c2_transform":
        comp = "C2"

    qi = state["quantizer"][level][orient]

    # These values evaulated in the loop definition in the spec, moving them
    # here saves a lot of computation
    ## Begin not in spec
    y1 = slice_top(state, sy, comp, level)
    y2 = slice_bottom(state, sy, comp, level)
    x1 = slice_left(state, sx, comp, level)
    x2 = slice_right(state, sx, comp, level)
    ## End not in spec

    ### for y in range(slice_top(state, sy,comp,level), slice_bottom(state, sy,comp,level)):
    ###     for x in range(slice_left(state, sx,comp,level), slice_right(state, sx,comp,level)):
    for y in range(y1, y2):  ## Not in spec
        for x in range(x1, x2):  ## Not in spec
            val = read_sintb(state)
            state[transform][level][orient][y][x] = inverse_quant(val, qi)


@ref_pseudocode
def color_diff_slice_band(state, level, orient, sx, sy):
    """(13.5.6.4)"""
    qi = state["quantizer"][level][orient]

    # These values evaulated in the loop definition in the spec, moving them
    # here saves a lot of computation
    ## Begin not in spec
    y1 = slice_top(state, sy, "C1", level)
    y2 = slice_bottom(state, sy, "C1", level)
    x1 = slice_left(state, sx, "C1", level)
    x2 = slice_right(state, sx, "C1", level)
    ## End not in spec

    ### for y in range(slice_top(state,sy,"C1",level), slice_bottom(state,sy,"C1",level)):
    ###     for x in range(slice_left(state,sx,"C1",level), slice_right(state,sx,"C1",level)):
    for y in range(y1, y2):  ## Not in spec
        for x in range(x1, x2):  ## Not in spec
            val = read_sintb(state)
            state["c1_transform"][level][orient][y][x] = inverse_quant(val, qi)
            val = read_sintb(state)
            state["c2_transform"][level][orient][y][x] = inverse_quant(val, qi)



"""
See `cli.py`. NB: This is a submodule to ensure that when this is executed as a
script, everything in that module is pickleable.
"""

import sys

from vc2_conformance.scripts.vc2_test_case_generator.cli import main


if __name__ == "__main__":
    sys.exit(main())



r"""
The :py:mod:`vc2_conformance.encoder` module implements a simple VC-2 encoder
which is used to produce test streams for conformance testing purposes (see
:py:mod:`vc2_conformance.test_cases`). It is extremely slow and performs only
simple picture compression, but is sufficiently flexible to support all VC-2
coding modes.

The encoder is principally concerned with carrying out the following tasks:

* Encoding the video and codec configuration into a sequence header
  (see :py:mod:`vc2_conformance.encoder.sequence_header`).
* Transforming, slicing and quantizing pictures (i.e. compressing them)
  (see :py:mod:`vc2_conformance.encoder.pictures`).
* Assembling sequences of data units comprising a complete VC-2 stream
  (see :py:mod:`vc2_conformance.encoder.sequence`).

This module does not generate serialised VC-2 bitstreams in binary format.
Instead, it generates a bitstream description data structure which may
subsequently be serialised by the bitstream serialiser in the
:py:mod:`vc2_conformance.bitstream` module. This design allows the generated
bitstream to be more easily manipulated prior to serialisation if required for
a particular test case.


Usage
-----

The encoder behaviour is controlled by a
:py:class:`~vc2_conformance.codec_features.CodecFeatures` dictionary. This
specifies picture/video format to be compressed (e.g. resolution etc.) along
with the coding options (e.g. wavelet transform and bitrate). See the
:py:mod:`vc2_conformance.codec_features` module for more details. There are
essentially two modes of operation, depending on the value of
:py:class:`~vc2_conformance.codec_features.CodecFeatures`\ ``["lossless"]``:

* Lossless mode: variable bitrate, qindex is always 0.
* Lossy mode: fixed bit rate, variable qindex.

A series of pictures may be encoded into a VC-2 sequence using the
:py:func:`vc2_conformance.encoder.make_sequence` function, and then serialised
into a binary bitstream as illustrated below::

    >>> # Define the video format to be encoded and basic coding options
    >>> from vc2_conformance.codec_features import CodecFeatures
    >>> codec_features = CodecFeatures(...)

    >>> # Encode a series of pictures
    >>> from vc2_conformance.encoder import make_sequence
    >>> pictures = [
    ...     {"Y": ..., "C1": ..., "C2": ..., "pic_num": 100},
    ...     # ...
    ... ]
    >>> sequence = make_sequence(codec_features, pictures)

    >>> # Serialise to a file
    >>> from vc2_conformance.bitstream import (
    ...     Stream,
    ...     autofill_and_serialise_stream,
    ... )
    >>> with open("bitstream.vc2", "wb") as f:
    ...     autofill_and_serialise_stream(f, Stream(sequences=[sequence]))


Bitstream conformance
---------------------

This encoder will produce bitstreams conformant with the VC-2 specification
whenever the coding parameters specified represent a legal combination.

In some cases, invalid coding options will cause the encoder to fail and raise
an exception in :py:mod:`vc2_conformance.encoder.exceptions`. For example, if
lossless coding and the low delay profile are requested simultaneously.

In other cases, the encoder will produce a bitstream, however this bitstream
will be non-conformant. For example, if the clean area defined is larger than
the frame size, the encoder will ignore this inconsistency of metadata and
produce a (non-conformant) bitstream anyway.

When conformant bitstreams are required, it is the responsibility of the user
of the encoder to ensure that the provided
:py:class:`~vc2_conformance.codec_features.CodecFeatures` are valid. In
practice, the easiest way to do this is to check for exceptions when the
encoder is used then use the bitstream validator
(:py:mod:`vc2_conformance.decoder`) to validate the generated bitstream.


Exceptions
----------

.. automodule:: vc2_conformance.encoder.exceptions


Sequence header generation
--------------------------

.. automodule:: vc2_conformance.encoder.sequence_header


Picture encoding & compression
------------------------------

.. automodule:: vc2_conformance.encoder.pictures


Sequence generation
-------------------

.. automodule:: vc2_conformance.encoder.sequence


Level constraints
-----------------

For the most part, all of the parameters which could be restricted by a VC-2
level are chosen in the supplied
:py:class:`~vc2_conformance.codec_features.CodecFeatures`. As such, choosing
parameters which comply with the declared level is the responsibility of the
caller (see comments above). However, some coding choices restricted by levels
are left up to this encoder, such as how video parameters are coded in a
sequence header. In these cases, the encoder makes choices which comply with
the supplied level, a process which may require a constraint solving procedure.

In principle level constaints, as expressed by constraints tables (see
:py:class:`vc2_conformance.constraint_table` and
:py:class:`vc2_conformance.level_constraints`), could require a full global
constraint solver to resolve. Fortunately, all existing VC-2 levels are
specified such that, once the level (and a few other parameters) have been
defined, almost all constrained parameters are independent meaning that global
constraint solving is not required. The only case where constraint dependencies
exist are the parameters relating to sequence headers. As a consequence the
:py:mod:`~vc2_conformance.encoder.sequence_header` generation module uses a
simple constraint solver internally.

.. note::

    The constraint parameter independence property of the VC-2 levels mentioned
    above is essential for the encoder to generate level-conforming bitstreams.
    A test in ``tests/encoder/test_level_constraints_assumptions.py`` is
    provided which will fail should a future VC-2 level not have this property.
    See the detailed documentation at the top of this file for a more thorough
    introduction and discussion of this topic.


"""

# Exceptions thrown when impossible requests are made of the encoder
from vc2_conformance.encoder.exceptions import *

# Sequence header construction
from vc2_conformance.encoder.sequence_header import *

# Picture compression
from vc2_conformance.encoder.pictures import *

# Complete sequence assembly
from vc2_conformance.encoder.sequence import *



r"""
.. _vc2-picture-explain:

``vc2-picture-explain``
=======================

A command-line utility which provides informative descriptions the raw video
format (see :ref:`file-format`) used by the VC-2 conformance software. As well
as providing an explanation of the format, where possible sample invocations of
`FFmpeg <https://ffmpeg.org/>`_ and `ImageMagick <https://imagemagick.org/>`_
are provided for viewing the raw files directly.

Example usage
-------------

An example invocation is shown below::

    $ vc2-picture-explain path/to/raw/picture_0.json
    Normative description
    =====================

    Picture coding mode: pictures_are_fields (1)

    Video parameters:

    * frame_width: 720
    * frame_height: 576
    * color_diff_format_index: color_4_2_2 (1)
    * source_sampling: interlaced (1)
    * top_field_first: True
    * frame_rate_numer: 25
    * frame_rate_denom: 1
    * pixel_aspect_ratio_numer: 12
    * pixel_aspect_ratio_denom: 11
    * clean_width: 704
    * clean_height: 576
    * left_offset: 8
    * top_offset: 0
    * luma_offset: 16
    * luma_excursion: 219
    * color_diff_offset: 128
    * color_diff_excursion: 224
    * color_primaries_index: sdtv_625 (2)
    * color_matrix_index: sdtv (1)
    * transfer_function_index: tv_gamma (0)

    Explanation (informative)
    =========================

    Each raw picture contains a single field. The top field comes first.

    Pictures contain three planar components: Y, Cb and Cr, in that order, which are
    4:2:2 subsampled.

    The Y component consists of 720x288 8 bit values. Expressible values run from 0
    (video level -0.07) to 255 (video level 1.09).

    The Cb and Cr components consist of 360x288 8 bit values. Expressible values run
    from 0 (video level -0.57) to 255 (video level 0.57).

    The color model uses the 'sdtv_625' primaries (ITU-R BT.601), the 'sdtv' color
    matrix (ITU-R BT.601) and the 'tv_gamma' transfer function (ITU-R BT.2020).

    The pixel aspect ratio is 12:11 (not to be confused with the frame aspect
    ratio).

    Example FFMPEG command (informative)
    ====================================

    The following command can be used to play back this video format using FFMPEG:

        $ ffplay \
            -f image2 \
            -video_size 720x288 \
            -framerate 50 \
            -pixel_format yuv422p \
            -i path/to/raw/picture_%d.raw \
            -vf weave=t,yadif,scale='trunc((iw*12)/11):ih'

    Where:

    * `-f image2` = Read pictures from individual files
    * `-video_size 720x288` = Picture size (not frame size).
    * `-framerate 50` = Picture rate (not frame rate)
    * `-pixel_format` = Specifies raw picture encoding.
    * `yuv` = Y C1 C2 color.
    * `422` = 4:2:2 color difference subsampling.
    * `p` = Planar format.
    * `-i path/to/raw/picture_%d.raw` = Input raw picture filename pattern
    * `-vf` = define a pipeline of video filtering operations
    * `weave=t` = interleave pairs of pictures, top field first
    * `yadif` = (optional) apply a deinterlacing filter for display purposes
    * `scale='trunc((iw*12)/11):ih'` = rescale non-square pixels for display with
      square pixels

    This command is provided as a minimal example for basic playback of this raw
    video format.  While it attempts to ensure correct frame rate, pixel aspect
    ratio, interlacing mode and basic pixel format, color model options are omitted
    due to inconsistent handling by FFMPEG.

    Example ImageMagick command (informative)
    =========================================

    The following command can be used to convert a single raw picture into PNG
    format for viewing in a conventional image viewer:

        $ convert \
            -size 720x288 \
            -depth 8 \
            -sampling-factor 4:2:2 \
            -interlace plane \
            -colorspace sRGB \
            yuv:path/to/raw/picture_0.raw \
            -resize 109.0909090909091%,100% \
            png24:path/to/raw/picture_0.raw

    Where:

    * `-size 720x288` = Picture size.
    * `-depth 8` = 8 bit values.
    * `-sampling-factor 4:2:2` = 4:2:2 color difference subsampling.
    * `-interlace plane` = Planar format (not related to video interlace mode).
    * `-colorspace sRGB` = Display as if using sRGB color.
    * `yuv:` = Input is Y C1 C2 picture.
    * `path/to/raw/picture_0.raw` = Input filename.
    * `-resize 109.0909090909091%,100%` = rescale non-square pixels for display with
      square pixels
    * `png24:path/to/raw/picture_0.png` = Save as 24-bit PNG (e.g. 8 bit channels)

    This command is provided as a minimal example for basic viewing of pictures.
    Interlacing and correct color model conversion are not implemented.


Arguments
---------

The complete set of arguments can be listed using ``--help``

.. program-output:: vc2-picture-explain --help

"""

import sys

from argparse import ArgumentParser

from fractions import Fraction

from vc2_data_tables import (
    SourceSamplingModes,
    PictureCodingModes,
    PRESET_COLOR_PRIMARIES,
    PresetColorMatrices,
    PRESET_COLOR_MATRICES,
    PRESET_TRANSFER_FUNCTIONS,
    ColorDifferenceSamplingFormats,
)

from vc2_conformance import __version__

from vc2_conformance.file_format import (
    get_metadata_and_picture_filenames,
    get_picture_filename_pattern,
    read_metadata,
)

from vc2_conformance.dimensions_and_depths import compute_dimensions_and_depths

from vc2_conformance.color_conversion import int_to_float

from vc2_conformance.py2x_compat import get_terminal_size
from vc2_conformance.string_utils import indent, wrap_paragraphs


def is_rgb_color(video_parameters):
    """
    Return True if the specified video parameters use RGB color. Returns False
    if a Luma+Chroma representation is used instead.
    """
    return video_parameters["color_matrix_index"] == PresetColorMatrices.rgb


PICTURE_COMPONENT_NAMES = {
    PresetColorMatrices.hdtv: ("Y", "Cb", "Cr"),
    PresetColorMatrices.sdtv: ("Y", "Cb", "Cr"),
    PresetColorMatrices.reversible: ("Y", "Cg", "Co"),
    PresetColorMatrices.rgb: ("G", "R", "B"),
    PresetColorMatrices.uhdtv: ("Y", "Cb", "Cr"),
}
"""
A lookup from :py:class:`~vc2_data_tables.PresetColorMatrices` to triples of
strings naming the components used by that format
"""

COLOR_DIFF_FORMAT_NAMES = {
    ColorDifferenceSamplingFormats.color_4_4_4: "4:4:4",
    ColorDifferenceSamplingFormats.color_4_2_2: "4:2:2",
    ColorDifferenceSamplingFormats.color_4_2_0: "4:2:0",
}
"""
Strings giving the normal human-readable name for a color difference sampling
format.
"""


def explain_interleaving(video_parameters, picture_coding_mode):
    """
    Produce a human-readable informative description of the source sampling and
    picture coding modes in use.
    """

    out = ""

    pictures_are_frames = picture_coding_mode == PictureCodingModes.pictures_are_frames
    progressive = video_parameters["source_sampling"] == SourceSamplingModes.progressive

    out += "Each raw picture contains a {}{}.".format(
        "whole frame" if pictures_are_frames else "single field",
        " (though the underlying video is {})".format(
            "progressive" if progressive else "interlaced"
        )
        if progressive != pictures_are_frames
        else "",
    )

    if not progressive or not pictures_are_frames:
        out += " The {} field comes first.".format(
            "top" if video_parameters["top_field_first"] else "bottom"
        )

    return out


def explain_component_order_and_sampling(video_parameters, picture_coding_mode):
    """
    Produce a human-readable informative definition of the component names,
    order and sampling mode.
    """
    y, c1, c2 = PICTURE_COMPONENT_NAMES[video_parameters["color_matrix_index"]]

    return (
        "Pictures contain three planar components: "
        "{}, {} and {}, in that order, which are {} subsampled."
    ).format(
        y,
        c1,
        c2,
        COLOR_DIFF_FORMAT_NAMES[video_parameters["color_diff_format_index"]],
    )


def explain_component_sizes(video_parameters, picture_coding_mode):
    """
    Produce a human-readable informative definition of the picture component
    sizes and bit depths.
    """
    out = ""

    dimensions_and_depths = compute_dimensions_and_depths(
        video_parameters,
        picture_coding_mode,
    )

    y_dd = dimensions_and_depths["Y"]
    c_dd = dimensions_and_depths["C1"]

    y_off_exc = (
        video_parameters["luma_offset"],
        video_parameters["luma_excursion"],
    )

    c_off_exc = (
        video_parameters["color_diff_offset"],
        video_parameters["color_diff_excursion"],
    )

    explanations = []

    if y_dd == c_dd and y_off_exc == c_off_exc:
        explanations.append(
            (
                "Each component consists of",
                y_dd,
                video_parameters["luma_offset"],
                video_parameters["luma_excursion"],
            )
        )
    else:
        y_name, c1_name, c2_name = PICTURE_COMPONENT_NAMES[
            video_parameters["color_matrix_index"]
        ]
        explanations.append(
            (
                "The {} component consists of".format(y_name),
                y_dd,
                video_parameters["luma_offset"],
                video_parameters["luma_excursion"],
            )
        )
        explanations.append(
            (
                "The {} and {} components consist of".format(c1_name, c2_name),
                c_dd,
                video_parameters["color_diff_offset"],
                video_parameters["color_diff_excursion"],
            )
        )

    for prefix, dd, offset, excursion in explanations:
        padding_bits = (dd.bytes_per_sample * 8) - dd.depth_bits
        out += (
            "{} {}x{} {} bit values{}{}. "
            "Expressible values run from 0 (video level {:0.2f}) "
            "to {} (video level {:0.2f})."
            "\n\n"
        ).format(
            prefix,
            dd.width,
            dd.height,
            dd.depth_bits,
            (
                (
                    " stored as {} bit ({} byte) values "
                    "(with the {} most significant bit{} set to 0)"
                ).format(
                    dd.bytes_per_sample * 8,
                    dd.bytes_per_sample,
                    padding_bits,
                    "s" if padding_bits != 1 else "",
                )
                if 8 * dd.bytes_per_sample != dd.depth_bits
                else ""
            ),
            " in little-endian byte order" if dd.bytes_per_sample > 1 else "",
            int_to_float(0, offset, excursion),
            (2 ** dd.depth_bits) - 1,
            int_to_float((2 ** dd.depth_bits) - 1, offset, excursion),
        )

    return out.rstrip()


def explain_color_model(video_parameters, picture_coding_mode):
    """
    Produce a human-readable informative definition of the color model in use.
    """
    primaries = PRESET_COLOR_PRIMARIES[video_parameters["color_primaries_index"]]
    matrix = PRESET_COLOR_MATRICES[video_parameters["color_matrix_index"]]
    transfer_function = PRESET_TRANSFER_FUNCTIONS[
        video_parameters["transfer_function_index"]
    ]

    out = (
        "The color model uses the '{}' primaries ({}), "
        "the '{}' color matrix ({}) "
        "and the '{}' transfer function ({})."
    ).format(
        primaries.name,
        primaries.specification,
        matrix.name,
        matrix.specification,
        transfer_function.name,
        transfer_function.specification,
    )

    return out.rstrip()


def explain_pixel_aspect_ratio(video_parameters, picture_coding_mode):
    """
    Produce a human-readable informative definition of the pixel aspect ratio.
    """
    return (
        "The pixel aspect ratio is {}:{} (not to be confused with the frame aspect ratio)."
    ).format(
        video_parameters["pixel_aspect_ratio_numer"],
        video_parameters["pixel_aspect_ratio_denom"],
    )


class CommandExplainer(object):
    """
    Build up a command line with notes explaining each part.

    Parameters
    ==========
    prefix : str
        Optionally add a prefix to the command (e.g. "$ ").

    Attributes
    ==========
    command : str
        The command.
    notes : [(start, end, note), ...]
        Notes relating to various substrings within the command.
        Non-overlapping.
    """

    def __init__(self, prefix=""):
        self.prefix = prefix
        self.command = ""

        # [(fragment, note), ...]
        self.notes = []

    def append(self, fragment, note=None, strip=True):
        """
        Append a fragment of text to the end of the command.

        Parameters
        ==========
        fragment : str
            The text to append
        note : str or None
            If given, a note explaining this fragment of the command.
        strip : bool
            If True (the default), the note will not apply to whitespace at
            either end of the fragment.
        """
        self.command += fragment

        if note is not None:
            if strip:
                self.notes.append((fragment.strip(), note))
            else:
                self.notes.append((fragment, note))

    def append_linebreak(self):
        self.append(" \\\n  ")

    def explain(self):
        """
        Return a markdown-formatted explanation of this command.
        """
        out = ""

        out += "    {}{}\n".format(
            self.prefix,
            indent(self.command, "    " + " " * len(self.prefix)).lstrip(),
        )

        # Add list of explanations
        if self.notes:
            out += "\n"
            out += "Where:\n"
            out += "\n"

            for (fragment, note) in self.notes:
                out += "* `{}` = {}\n".format(
                    fragment,
                    note,
                )

        out = out.rstrip()

        return out

    def __repr__(self):
        return "<{!r} {!r}>".format(
            type(self).__name__,
            self.command,
        )


class UnsupportedPictureFormat(Exception):
    """
    Thrown when a picture format is not supported by a command line tool.
    """


def example_ffmpeg_command(picture_filename, video_parameters, picture_coding_mode):
    """
    Create a :py:class:`CommandExplainer` defining an FFmpeg ``ffplay`` command
    to play the specified video format.

    The resulting commands attempt to achieve the correct:

    * Resolution
    * Bit depth
    * Pixel aspect ratio
    * Color subsampling mode
    * Frame rate
    * Interlacing (fields are always converted into frames and, if the
      underlying picture is interlaced, deinterlacing is applied)
    * RGB vs Luma + Chroma representation

    In some cases, :py:exc:`UnsupportedPictureFormat` will be raised where the
    format is sufficiently unusual as to be completely unsupported by FFMPEG
    (e.g.  extremely high bit depths or color subsampling under RGB mode).

    Color model parameters are essentially ignored.

    Finally, the 'clean area' parameters are ignored completely.
    """
    command = CommandExplainer("$ ")

    command.append("ffplay")

    # ------------------------------------------------------------------------
    # Set input image format
    # ------------------------------------------------------------------------

    command.append_linebreak()
    command.append("-f image2", "Read pictures from individual files")

    # ------------------------------------------------------------------------
    # Set picture size
    # ------------------------------------------------------------------------

    dimensions_and_depths = compute_dimensions_and_depths(
        video_parameters,
        picture_coding_mode,
    )

    command.append_linebreak()
    command.append(
        "-video_size {}x{}".format(
            dimensions_and_depths["Y"].width,
            dimensions_and_depths["Y"].height,
        ),
        "Picture size (not frame size).",
    )

    # ------------------------------------------------------------------------
    # Set frame rate
    # ------------------------------------------------------------------------

    frame_rate = Fraction(
        video_parameters["frame_rate_numer"], video_parameters["frame_rate_denom"]
    )

    command.append_linebreak()
    command.append(
        "-framerate {}".format(
            frame_rate
            if picture_coding_mode == PictureCodingModes.pictures_are_frames
            else frame_rate * 2
        ),
        "Picture rate (not frame rate)",
    )

    # ------------------------------------------------------------------------
    # Set pixel format (e.g. RGB-vs-YCbCr, bit depth, color subsampling)
    # ------------------------------------------------------------------------

    command.append_linebreak()
    command.append("-pixel_format ", "Specifies raw picture encoding.")

    if is_rgb_color(video_parameters):
        command.append("gbr", "RGB color.")

        if (
            video_parameters["color_diff_format_index"]
            != ColorDifferenceSamplingFormats.color_4_4_4
        ):
            raise UnsupportedPictureFormat(
                "Only 4:4:4 color difference sampling is supported for RGB video."
            )
    elif video_parameters["color_matrix_index"] == PresetColorMatrices.reversible:
        raise UnsupportedPictureFormat(
            "Y Cg Co color ('reversible' color matrix) unsupported."
        )
    else:
        command.append("yuv", "Y C1 C2 color.")

        color_difference_sampling_format = COLOR_DIFF_FORMAT_NAMES[
            video_parameters["color_diff_format_index"]
        ]
        command.append(
            color_difference_sampling_format.replace(":", ""),
            "{} color difference subsampling.".format(
                color_difference_sampling_format,
            ),
        )

    command.append("p", "Planar format.")

    if (
        dimensions_and_depths["Y"].bytes_per_sample
        != dimensions_and_depths["C1"].bytes_per_sample
    ):
        raise UnsupportedPictureFormat(
            "Luma and color difference components differ in depth ({} bits and {} bits)".format(
                dimensions_and_depths["Y"].bytes_per_sample,
                dimensions_and_depths["C1"].bytes_per_sample,
            )
        )

    depth_bits = dimensions_and_depths["Y"].depth_bits
    if depth_bits == 8:
        # No extension required for 8-bit values
        pass
    elif depth_bits in (9, 10, 12, 14, 16):
        command.append(
            "{}le".format(depth_bits),
            "{} bit little-endian values, LSB-aligned within 16 bit words.".format(
                depth_bits
            ),
        )
    else:
        raise UnsupportedPictureFormat(
            "Unsupported bit depth: {} bits".format(
                dimensions_and_depths["Y"].depth_bits
            )
        )

    # ------------------------------------------------------------------------
    # Input file format
    # ------------------------------------------------------------------------

    command.append_linebreak()
    command.append(
        "-i {}".format(get_picture_filename_pattern(picture_filename)),
        "Input raw picture filename pattern",
    )

    # ------------------------------------------------------------------------
    # Frames-from-Fields (de)interlacing
    # ------------------------------------------------------------------------

    vf_args = []

    if picture_coding_mode == PictureCodingModes.pictures_are_fields:
        # When pictures are fields we (at least) need to interleave pairs of
        # pictures into whole frames. For truly interlaced content, a
        # deinterlace filter can also be applied for clean display.
        vf_args.append(
            (
                "weave={}".format(
                    "t" if video_parameters["top_field_first"] else "b",
                ),
                "interleave pairs of pictures, {} field first".format(
                    "top" if video_parameters["top_field_first"] else "bottom",
                ),
            )
        )

        if video_parameters["source_sampling"] == SourceSamplingModes.interlaced:
            vf_args.append((",",))
            vf_args.append(
                (
                    "yadif",
                    "(optional) apply a deinterlacing filter for display purposes",
                )
            )
    elif (
        picture_coding_mode == PictureCodingModes.pictures_are_frames
        and video_parameters["source_sampling"] == SourceSamplingModes.interlaced
    ):
        # When pictures are frames containing interlaced content, we can
        # optionally apply a deinterlace filter
        vf_args.append(
            (
                "setfield={}".format(
                    "tff" if video_parameters["top_field_first"] else "bff",
                ),
                "define the field order as {} field first".format(
                    "top" if video_parameters["top_field_first"] else "bottom",
                ),
            )
        )

        vf_args.append((",",))
        vf_args.append(
            ("yadif", "(optional) apply a deinterlacing filter for display purposes")
        )

    # ------------------------------------------------------------------------
    # Pixel aspect ratio
    # ------------------------------------------------------------------------
    pixel_aspect_ratio = Fraction(
        video_parameters["pixel_aspect_ratio_numer"],
        video_parameters["pixel_aspect_ratio_denom"],
    )
    if pixel_aspect_ratio != 1:
        if vf_args:
            vf_args.append((",",))
        vf_args.append(
            (
                "scale='trunc((iw*{})/{}):ih'".format(
                    pixel_aspect_ratio.numerator,
                    pixel_aspect_ratio.denominator,
                ),
                "rescale non-square pixels for display with square pixels",
            )
        )

    if vf_args:
        command.append_linebreak()
        command.append("-vf ", "define a pipeline of video filtering operations")
        for args in vf_args:
            command.append(*args)

    return command


def example_imagemagick_command(
    picture_filename, video_parameters, picture_coding_mode
):
    """
    Create a :py:class:`CommandExplainer` defining an ImagMagick ``display``
    command to display a single picture in the raw video format.

    The resulting commands attempt to achieve the correct:

    * Resolution
    * Color subsampling mode
    * RGB vs Luma + Color Difference representation

    Unfortunately, bit depths greater than 8 bits are unsupported due to
    the unreliability of ImageMagick's handling of 16-bit color.

    In some cases, :py:exc:`UnsupportedPictureFormat` will be raised where the
    format is sufficiently unusual as to be completely unsupported by
    ImageMagick (e.g.  extremely high bit depths or color subsampling under
    RGB mode).

    Color model parameters are essentially ignored.

    Interlaced pictures (or progressive frames encoded as fields) are not
    (de)interlaced: pictures are shown as-is.

    Finally, the 'clean area' parameters are ignored completely.
    """
    command = CommandExplainer("$ ")

    command.append("convert")

    # ------------------------------------------------------------------------
    # Set picture size
    # ------------------------------------------------------------------------

    dimensions_and_depths = compute_dimensions_and_depths(
        video_parameters,
        picture_coding_mode,
    )

    command.append_linebreak()
    command.append(
        "-size {}x{}".format(
            dimensions_and_depths["Y"].width,
            dimensions_and_depths["Y"].height,
        ),
        "Picture size.",
    )

    # ------------------------------------------------------------------------
    # Depth
    # ------------------------------------------------------------------------

    if dimensions_and_depths["Y"].depth_bits != dimensions_and_depths["C1"].depth_bits:
        raise UnsupportedPictureFormat(
            "Luma and chroma components differ in depth ({} bits and {} bits)".format(
                dimensions_and_depths["Y"].depth_bits,
                dimensions_and_depths["C1"].depth_bits,
            )
        )

    depth_bits = dimensions_and_depths["Y"].depth_bits
    if depth_bits == 8:
        command.append_linebreak()
        command.append(
            "-depth {}".format(depth_bits),
            "{} bit values.".format(depth_bits),
        )
    else:
        raise UnsupportedPictureFormat(
            "Unsupported bit depth: {} bits".format(
                dimensions_and_depths["Y"].depth_bits
            )
        )

    # ------------------------------------------------------------------------
    # Sampling factor
    # ------------------------------------------------------------------------

    if is_rgb_color(video_parameters):
        if (
            video_parameters["color_diff_format_index"]
            != ColorDifferenceSamplingFormats.color_4_4_4
        ):
            raise UnsupportedPictureFormat(
                "Only 4:4:4 color difference sampling is supported for RGB video."
            )
    elif video_parameters["color_matrix_index"] == PresetColorMatrices.reversible:
        raise UnsupportedPictureFormat(
            "Y Cg Co color ('reversible' color matrix) unsupported."
        )
    else:
        color_difference_sampling_format = COLOR_DIFF_FORMAT_NAMES[
            video_parameters["color_diff_format_index"]
        ]
        command.append_linebreak()
        command.append(
            "-sampling-factor {}".format(color_difference_sampling_format),
            "{} color difference subsampling.".format(
                color_difference_sampling_format,
            ),
        )

    # ------------------------------------------------------------------------
    # Planar video format
    # ------------------------------------------------------------------------

    command.append_linebreak()
    command.append(
        "-interlace plane", "Planar format (not related to video interlace mode)."
    )

    # ------------------------------------------------------------------------
    # Colorspace
    # ------------------------------------------------------------------------

    command.append_linebreak()
    command.append("-colorspace sRGB", "Display as if using sRGB color.")

    # ------------------------------------------------------------------------
    # Input file
    # ------------------------------------------------------------------------

    command.append_linebreak()
    if is_rgb_color(video_parameters):
        command.append(
            "rgb:",
            "Input is RGB picture.",
        )
    else:
        command.append(
            "yuv:",
            "Input is Y C1 C2 picture.",
        )

    command.append(picture_filename, "Input filename.")

    # ------------------------------------------------------------------------
    # Convert GBR to RGB channel order expected by ImageMagick
    # ------------------------------------------------------------------------

    if is_rgb_color(video_parameters):
        command.append_linebreak()
        command.append(
            "-separate -swap 1,2 -swap 0,1 -combine",
            "Change from GRB to RGB channel order (expected by ImageMagick).",
        )

    # ------------------------------------------------------------------------
    # Pixel aspect ratio
    # ------------------------------------------------------------------------
    pixel_aspect_ratio = Fraction(
        video_parameters["pixel_aspect_ratio_numer"],
        video_parameters["pixel_aspect_ratio_denom"],
    )
    if pixel_aspect_ratio != 1:
        command.append_linebreak()
        command.append(
            "-resize {}%,100%".format(float(pixel_aspect_ratio * 100)),
            "rescale non-square pixels for display with square pixels",
        )

    # ------------------------------------------------------------------------
    # Output file (PNG)
    # ------------------------------------------------------------------------

    command.append_linebreak()
    command.append(
        "png24:{}png".format(picture_filename[:-3]),
        "Save as 24-bit PNG (e.g. 8 bit channels)",
    )

    return command


def explain_ffmpeg_command(picture_filename, video_parameters, picture_coding_mode):
    """
    Produce a human-readable informative message explaining how FFMPEG can be
    used to playback the raw video format.
    """
    try:
        command = example_ffmpeg_command(
            picture_filename, video_parameters, picture_coding_mode
        )
    except UnsupportedPictureFormat as e:
        return "No FFMPEG command is available for this raw video format ({}).".format(
            e
        )

    out = ""

    out += (
        "The following command can be used to play back this video format using FFMPEG:"
    )

    out += "\n"
    out += "\n"

    out += command.explain()

    out += "\n"
    out += "\n"

    out += "This command is provided as a minimal example for basic playback "
    out += "of this raw video format.  While it attempts to ensure correct "
    out += "frame rate, pixel aspect ratio, interlacing mode and basic pixel "
    out += "format, color model options are omitted due to inconsistent "
    out += "handling by FFMPEG."

    return out


def explain_imagemagick_command(
    picture_filename, video_parameters, picture_coding_mode
):
    """
    Produce a human-readable informative message explaining how ImageMagick can be
    used to display a picture in the raw format.
    """
    try:
        command = example_imagemagick_command(
            picture_filename, video_parameters, picture_coding_mode
        )
    except UnsupportedPictureFormat as e:
        return "No ImageMagick command is available for this raw picture format ({}).".format(
            e
        )

    out = ""

    out += "The following command can be used to convert a single raw picture "
    out += "into PNG format for viewing in a conventional image viewer:"

    out += "\n"
    out += "\n"

    out += command.explain()

    out += "\n"
    out += "\n"

    out += "This command is provided as a minimal example for basic viewing "
    out += "of pictures. Interlacing and correct color model conversion are "
    out += "not implemented."

    return out


def main(*args, **kwargs):
    parser = ArgumentParser(
        description="""
        Informatively explain the video format used a raw video
        file generated by the VC-2 conformance software.
    """
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__),
    )

    parser.add_argument(
        "filename",
        help="""
            The filename of a .raw or .json raw video file.
        """,
    )

    args = parser.parse_args(*args, **kwargs)

    metadata_filename, picture_filename = get_metadata_and_picture_filenames(
        args.filename
    )

    with open(metadata_filename, "rb") as f:
        (
            video_parameters,
            picture_coding_mode,
            picture_number,
        ) = read_metadata(f)

    out = ""

    out += "Normative description\n"
    out += "=====================\n"
    out += "\n"
    out += "Picture coding mode: {} ({:d})\n".format(
        picture_coding_mode.name,
        picture_coding_mode,
    )
    out += "\n"
    out += "Video parameters:\n"
    out += "\n"
    out += str(video_parameters).partition("\n")[2].replace("  ", "* ") + "\n"
    out += "\n"
    out += "Explanation (informative)\n"
    out += "=========================\n"
    out += "\n"
    out += explain_interleaving(video_parameters, picture_coding_mode) + "\n"
    out += "\n"
    out += (
        explain_component_order_and_sampling(video_parameters, picture_coding_mode)
        + "\n"
    )
    out += "\n"
    out += explain_component_sizes(video_parameters, picture_coding_mode) + "\n"
    out += "\n"
    out += explain_color_model(video_parameters, picture_coding_mode) + "\n"
    out += "\n"
    out += explain_pixel_aspect_ratio(video_parameters, picture_coding_mode) + "\n"
    out += "\n"
    out += "Example FFMPEG command (informative)\n"
    out += "====================================\n"
    out += "\n"
    out += (
        explain_ffmpeg_command(picture_filename, video_parameters, picture_coding_mode)
        + "\n"
    )
    out += "\n"
    out += "Example ImageMagick command (informative)\n"
    out += "=========================================\n"
    out += "\n"
    out += (
        explain_imagemagick_command(
            picture_filename, video_parameters, picture_coding_mode
        )
        + "\n"
    )

    width = get_terminal_size()[0]
    print(wrap_paragraphs(out, width))

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
The :py:mod:`vc2_conformance.file_format` module contains functions for reading
and writing raw pictures and their metadata as files.

Picture and metadata files must come in pairs with the raw planar picture data
file having the extension '.raw' and the metadata file having the extension
'.json'. See :ref:`file-format` for a description of the file format.

..
    NB: If you're reading this the 'file-format' reference above probably isn't
    clickable :). It points to the documentation in
    ``docs/source/user_guide/file_format.rst``.

The following functions may be used to read and write picture/metadata files:

.. autofunction:: read

.. autofunction:: write

The above functions are just wrappers around the following functions which read
and write picture and metadata files in isolation:

.. autofunction:: read_metadata

.. autofunction:: read_picture

.. autofunction:: write_metadata

.. autofunction:: write_picture

Finally, the following function may be used to get the filenames for both parts
of a picture/metadata file pair:

.. autofunction:: get_metadata_and_picture_filenames

"""

import os
import re
import json

import numpy as np

from vc2_data_tables import (
    ColorDifferenceSamplingFormats,
    PictureCodingModes,
    SourceSamplingModes,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
)

from vc2_conformance.pseudocode.video_parameters import VideoParameters

from vc2_conformance.dimensions_and_depths import compute_dimensions_and_depths


__all__ = [
    "read",
    "write",
    "get_metadata_and_picture_filenames",
    "read_metadata",
    "read_picture",
    "write_metadata",
    "write_picture",
]


def get_metadata_and_picture_filenames(filename):
    """
    Given either the filename of a saved picture (.raw) or metadata file
    (.json), return a (metadata_filename, picture_filename) tuple with the
    names of the two corresponding files.
    """
    base_name = os.path.splitext(filename)[0]
    return (
        "{}.json".format(base_name),
        "{}.raw".format(base_name),
    )


def get_picture_filename_pattern(filename):
    """
    Given the filename of a picture file (*.raw), return a version with the
    number replaced with '%d'.
    """
    return re.sub(
        r"(.*)_[0-9]+\.raw",
        r"\1_%d.raw",
        filename,
    )


def write(picture, video_parameters, picture_coding_mode, filename):
    """
    Write a picture to a data and metadata file.

    Convenience wrapper around :py:func:`write_picture` and
    :py:func:`write_metadata`.

    Parameters
    ==========
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    filename : str
        The filename of either the picture data file (.raw) or metadata file
        (.json). The name of the other file will be inferred automatically.
    """
    metadata_filename, picture_filename = get_metadata_and_picture_filenames(filename)

    with open(metadata_filename, "wb") as f:
        write_metadata(picture, video_parameters, picture_coding_mode, f)

    with open(picture_filename, "wb") as f:
        write_picture(picture, video_parameters, picture_coding_mode, f)


def read(filename):
    """
    Read a picture from a data and metadata file.

    Convenience wrapper around :py:func:`read_picture` and
    :py:func:`read_metadata`.

    Parameters
    ==========
    filename : str
        The filename of either the picture data file (.raw) or metadata file
        (.json). The name of the other file will be inferred automatically.

    Returns
    =======
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    """
    metadata_filename, picture_filename = get_metadata_and_picture_filenames(filename)

    with open(metadata_filename, "rb") as f:
        (
            video_parameters,
            picture_coding_mode,
            picture_number,
        ) = read_metadata(f)

    with open(picture_filename, "rb") as f:
        picture = read_picture(
            video_parameters,
            picture_coding_mode,
            picture_number,
            f,
        )

    return (picture, video_parameters, picture_coding_mode)


def write_picture(picture, video_parameters, picture_coding_mode, file):
    """
    Write a decoded picture to a file as a planar image.

    Parameters
    ==========
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    file : :py:class:`file`
        A file open for binary writing.
    """
    dims_and_depths = compute_dimensions_and_depths(
        video_parameters, picture_coding_mode
    )

    for (
        component,
        (width, height, depth_bits, bytes_per_sample),
    ) in dims_and_depths.items():
        # We use native Python integers in a dtype=object array to ensure we
        # can support arbitrary bit depths.
        #
        # NB: we make a copy as we will mutate later.
        values = np.array(picture[component], dtype=object)

        # Write as little-endian representation (NB: this rather explicit
        # expansion supports arbitrary depth values beyond those natively
        # supported by Numpy).
        out = np.zeros((height, width, bytes_per_sample), dtype=np.uint8)
        for byte in range(bytes_per_sample):
            out[:, :, byte] = values & 0xFF
            if byte != bytes_per_sample - 1:
                values >>= 8

        file.write(out.tobytes())


def write_metadata(picture, video_parameters, picture_coding_mode, file):
    """
    Write the metadata associated with a decoded picture to a file as JSON.

    Parameters
    ==========
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    file : :py:class:`file`
        A file open for binary writing.
    """

    # Conversion below is necessary under Python 2.x where IntEnum values are
    # not correctly serialised as integers but instead into invalid JSON.
    file.write(
        json.dumps(
            {
                "video_parameters": {
                    key: int(value)
                    if (isinstance(value, int) and not isinstance(value, bool))
                    else value
                    for key, value in video_parameters.items()
                },
                "picture_coding_mode": int(picture_coding_mode),
                "picture_number": str(picture["pic_num"]),
            }
        ).encode("utf-8")
    )


def read_metadata(file):
    """
    Read a JSON picture metadata file.

    Parameters
    ==========
    file : :py:class:`file`
        A file open for binary reading.

    Returns
    =======
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    picture_number : int
    """
    metadata = json.loads(file.read().decode("utf-8"))

    video_parameters = VideoParameters(metadata["video_parameters"])

    # Convert back into native types
    for name, int_enum_type in [
        ("color_diff_format_index", ColorDifferenceSamplingFormats),
        ("source_sampling", SourceSamplingModes),
        ("color_primaries_index", PresetColorPrimaries),
        ("color_matrix_index", PresetColorMatrices),
        ("transfer_function_index", PresetTransferFunctions),
    ]:
        video_parameters[name] = int_enum_type(video_parameters[name])

    picture_coding_mode = PictureCodingModes(metadata["picture_coding_mode"])

    picture_number = int(metadata["picture_number"])

    return (video_parameters, picture_coding_mode, picture_number)


def read_picture(video_parameters, picture_coding_mode, picture_number, file):
    """
    Read a picture from a file.

    Parameters
    ==========
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    picture_number : int
    file : :py:class:`file`
        A file open for binary reading.

    Returns
    =======
    picture : {"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}
    """
    picture = {"pic_num": picture_number}

    dims_and_depths = compute_dimensions_and_depths(
        video_parameters, picture_coding_mode
    )
    for (
        component,
        (width, height, depth_bits, bytes_per_sample),
    ) in dims_and_depths.items():
        data = np.frombuffer(
            file.read(height * width * bytes_per_sample),
            dtype=np.uint8,
        ).reshape(height, width, bytes_per_sample)

        # Mask off just the intended bits
        msb_byte = ((depth_bits + 7) // 8) - 1
        if msb_byte < bytes_per_sample:
            data = np.require(data, requirements="W")
            data[:, :, msb_byte + 1 :] = 0
        if depth_bits % 8 != 0:
            data = np.require(data, requirements="W")
            data[:, :, msb_byte] &= (1 << (depth_bits % 8)) - 1

        # We use a dtype=object array so that we can use Python's arbitrary
        # precision integers in order to support arbitrary bit depths
        values = np.zeros((height, width), dtype=object)

        # Read little endian (NB, this method supports arbitrary width values)
        for byte in reversed(range(bytes_per_sample)):
            if byte != bytes_per_sample - 1:
                values <<= 8
            values += data[:, :, byte]

        picture[component] = values.tolist()

    return picture
